# -*- coding: utf-8 -*-
"""
Summary
-------

Functions to compare and plot comparison of :class:`~radis.spectrum.spectrum.Spectrum`
objects


Routine Listings
----------------

- :func:`~radis.spectrum.compare.get_diff`
- :func:`~radis.spectrum.compare.get_distance`
- :func:`~radis.spectrum.compare.get_ratio`
- :func:`~radis.spectrum.compare.get_residual`
- :func:`~radis.spectrum.compare.get_residual_integral`
- :func:`~radis.spectrum.compare.plot_diff`


-------------------------------------------------------------------------------



"""

from warnings import warn

import numpy as np

from radis.misc.arrays import array_allclose
from radis.misc.basics import compare_dict, compare_lists
from radis.misc.curve import curve_distance, curve_divide, curve_substract
from radis.spectrum.spectrum import Spectrum
from radis.spectrum.utils import cast_waveunit, format_xlabel, make_up, make_up_unit

# %% ======================================================================
# External functions
# ----------------
# XXX =====================================================================


def get_diff(
    s1: Spectrum,
    s2: Spectrum,
    var,
    wunit="default",
    Iunit="default",
    resample=True,
    diff_window=0,
):
    # type: (Spectrum, Spectrum, str, str, str, str, bool, int) -> np.array, np.array
    """Get the difference between 2 spectra.

    Basically returns w1, I1 - I2 where (w1, I1) and (w2, I2) are the values of
    s1 and s2 for variable var. (w2, I2) is linearly interpolated if needed.

        .. math::

            dI = I_1 - I_2


    Parameters
    ----------
    s1, s2: Spectrum objects
        2 spectra to compare.
    var: str
        spectral quantity (ex: ``'radiance'``, ``'transmittance'``...)
    wunit: ``'nm'``, ``'cm-1'``, ``'nm_vac'``
        waveunit to compare in: wavelength air, wavenumber, wavelength vacuum
    Iunit: str
        if ``'default'`` use s1 unit for variable var
    medium: 'air', 'vacuum', default'
        propagating medium to compare in (if in wavelength)

    Other Parameters
    ----------------
    resample: bool
        if not ``True``, wavelength must be equals. Else, resample ``s2`` on
        ``s1`` if needed.
    diff_window: int
        If non 0, calculates diff by offsetting s1 by ``diff_window`` number of
        units on either side, and returns the minimum. Compensates for experimental
        errors on the w axis. Default 0. (look up code for more details...)

    Returns
    -------
    w1, Idiff: array
        difference interpolated on the wavespace range of the first Spectrum

    Notes
    -----

    Uses :func:`~radis.misc.curve.curve_substract` internally

    See Also
    --------

    :func:`~radis.spectrum.compare.get_ratio`,
    :func:`~radis.spectrum.compare.get_distance`,
    :func:`~radis.spectrum.compare.get_residual`,
    :func:`~radis.spectrum.compare.get_residual_integral`,
    :func:`~radis.spectrum.compare.plot_diff`,
    :meth:`~radis.spectrum.spectrum.compare_with`
    """
    # TODO : ensure objects are sorted and if so; forward to curve_substract()

    var, wunit, Iunit = get_default_units(s1, s2, var=var, wunit=wunit, Iunit=Iunit)

    # Get data
    # ----
    w1, I1 = s1.get(var, wunit=wunit, Iunit=Iunit, trim_nan=True)
    w2, I2 = s2.get(var, wunit=wunit, Iunit=Iunit, trim_nan=True)

    if not resample:
        if not array_allclose(w1, w2):
            raise AssertionError("Wavespace are not the same: use Spectrum.resample()")

    # Compute : I1 - I2 (on same range)
    w1, Idiff = curve_substract(w1, I1, w2, I2)
    if diff_window:
        # allow fluctuation from diff_window units. Kinda compensates
        # for experimental errors on x-axis
        diff_list = []
        I2_interp = I1 - Idiff
        for i in range(-diff_window, diff_window + 1):
            diff_list.append(I2_interp - np.roll(I1, i))
        # get minimum in abs value
        diff_list = np.array(diff_list)
        b = np.abs(diff_list).argmin(axis=0)
        Idiff = np.choose(b, diff_list)  # keeps sign
    return w1, Idiff


def get_ratio(
    s1: Spectrum, s2: Spectrum, var, wunit="default", Iunit="default", resample=True
):
    # type: (Spectrum, Spectrum, str, str, str, str, bool) -> np.array, np.array
    """Get the ratio between two spectra Basically returns w1, I1 / I2 where
    (w1, I1) and (w2, I2) are the values of s1 and s2 for variable var. (w2,
    I2) is linearly interpolated if needed.

        .. math::

            R = I_1 / I_2


    Parameters
    ----------
    s1, s2: Spectrum objects
        :py:class:`~radis.spectrum.spectrum.Spectrum`
    var: str
        spectral quantity
    wunit: ``'nm'``, ``'cm-1'``, ``'nm_vac'``
        waveunit to compare in: wavelength air, wavenumber, wavelength vacuum
    Iunit: str
        if ``'default'`` use s1 unit for variable var

    Notes
    -----
    Uses :func:`~radis.misc.curve.curve_divide` internally

    See Also
    --------

    :func:`~radis.spectrum.compare.get_diff`,
    :func:`~radis.spectrum.compare.get_distance`,
    :func:`~radis.spectrum.compare.get_residual`,
    :func:`~radis.spectrum.compare.get_residual_integral`,
    :func:`~radis.spectrum.compare.plot_diff`,
    :meth:`~radis.spectrum.spectrum.compare_with`
    """

    var, wunit, Iunit = get_default_units(s1, s2, var=var, wunit=wunit, Iunit=Iunit)

    # Get data
    # ----
    w1, I1 = s1.get(var, wunit=wunit, Iunit=Iunit)
    w2, I2 = s2.get(var, wunit=wunit, Iunit=Iunit)

    if not resample:
        if not array_allclose(w1, w2):
            raise AssertionError("Wavespace are not the same: use Spectrum.resample()")

    # Compute: I1 / I2 (on same range)
    return curve_divide(w1, I1, w2, I2)


def get_distance(
    s1: Spectrum, s2: Spectrum, var, wunit="default", Iunit="default", resample=True
):
    # type: (Spectrum, Spectrum, str, str, str, str, bool) -> np.array, np.array
    """Get a regularized Euclidian distance between two spectra ``s1`` and
    ``s2``

    This regularized Euclidian distance minimizes the effect of a small shift in
    wavelength between the two spectra

    .. math::

        D(w_1)[i] = \sqrt{ \sum_j (\hat{I_1}[i]  - \hat{I_2}[j] )^2 + (\hat{w_1}[i] - \hat{w_2}[j])^2}

    Where values are normalized as:

    .. math::

        \hat{A} = \\frac{A}{max(A) - min(A)}

    If waveranges dont match, ``s2`` is interpolated over ``s1``.

    .. warning ::

        This is a distance on both the waverange and the intensity axis.
        It may be used to compensate for a small offset in your experimental
        spectrum (due to wavelength calibration, for instance) but can lead
        to wrong fits easily. Plus, it is very cost-intensive! Better use
        :func:`~radis.spectrum.compare.get_residual` for an automatized procedure.


    Parameters
    ----------
    s1, s2: Spectrum objects
        :py:class:`~radis.spectrum.spectrum.Spectrum`
    var: str
        spectral quantity
    wunit: ``'nm'``, ``'cm-1'``, ``'nm_vac'``
        waveunit to compare in: wavelength air, wavenumber, wavelength vacuum
    Iunit: str
        if ``'default'`` use s1 unit for variable var
    medium: 'air', 'vacuum', default'
        propagating medium to compare in (if in wavelength)

    Notes
    -----
    Uses :func:`~radis.misc.curve.curve_distance` internally

    See Also
    --------
    :func:`~radis.spectrum.compare.get_diff`,
    :func:`~radis.spectrum.compare.get_ratio`,
    :func:`~radis.spectrum.compare.get_residual`,
    :func:`~radis.spectrum.compare.get_residual_integral`,
    :func:`~radis.spectrum.compare.plot_diff`,
    :meth:`~radis.spectrum.spectrum.compare_with`
    """
    # TODO: normalize with Imax, wmax

    var, wunit, Iunit = get_default_units(s1, s2, var=var, wunit=wunit, Iunit=Iunit)

    # Get data
    # ----
    w1, I1 = s1.get(var, wunit=wunit, Iunit=Iunit)
    w2, I2 = s2.get(var, wunit=wunit, Iunit=Iunit)

    if not resample:
        if not array_allclose(w1, w2):
            raise AssertionError("Wavespace are not the same: use Spectrum.resample()")

    # Compute: euclidian distance from w1, I1
    return curve_distance(w1, I1, w2, I2, discard_out_of_bounds=True)


def get_residual(
    s1: Spectrum,
    s2: Spectrum,
    var,
    norm="L2",
    ignore_nan=False,
    diff_window=0,
    normalize=False,
    normalize_how="max",
    wunit="default",
    Iunit="default",
) -> float:
    # type: (Spectrum, Spectrum, str, bool, int) -> np.array, np.array
    """Returns L2 norm of ``s1`` and ``s2``

    For ``I1``, ``I2``, the values of variable ``var`` in ``s1`` and ``s2``,
    respectively, residual is calculated as:

    For ``L2`` norm:

        .. math::

            res = \\frac{\\sqrt{\\sum_i {(s_1[i]-s_2[i])^2}}}{N}.

    For ``L1`` norm:

        .. math::

            res = \\frac{\\sqrt{\\sum_i {|s_1[i]-s_2[i]|}}}{N}.


    Parameters
    ----------
    s1, s2: :class:`~radis.spectrum.spectrum.Spectrum` objects
        if not on the same range, ``s2`` is resampled on ``s1``.
    var: str
        spectral array
    norm: 'L2', 'L1'
        which norm to use

    Other Parameters
    ----------------
    ignore_nan: boolean
        if ``True``, ignore ``nan`` in the difference between s1 and s2 (ex: out of bound)
        when calculating residual. Default ``False``. Note: ``get_residual`` will still
        fail if there are ``nan`` in initial Spectrum.
    normalize: bool, or tuple
        if ``True``, normalize the two spectra before calculating the residual.
        If a tuple (ex: ``(4168, 4180)``), normalize on this range only. The unit
        is that of the first Spectrum by default (use ``wunit`` to change). Ex::

            s_exp   # in 'nm'
            s_calc  # in 'cm-1'
            get_residual(s_exp, s_calc, normalize=(4178, 4180))  # interpreted as 'nm'

    normalize_how: ``'max'``, ``'area'``, ``'mean'``
        how to normalize. ``'max'`` is the default but may not be suited for very
        noisy experimental spectra. ``'area'`` will normalize the integral to 1.
        ``'mean'`` will normalize by the mean amplitude value
    wunit: str
        used if normalized is a range. If ``'default'``, use first spectrum unit.
    Iunit: str
        if ``'default'``, use first spectrum unit

    Notes
    -----
    0 values for I1 yield nans except if I2 = I1 = 0

    when s1 and s2 dont have the size wavespace range, they are automatically
    resampled through get_diff on 's1' range

    Implementation of ``L2`` norm::

        np.sqrt((dI**2).sum())/len(dI)

    Implementation of ``L1`` norm::

        np.abs(dI).sum()/len(dI)

    See Also
    --------
    :func:`~radis.spectrum.compare.get_diff`,
    :func:`~radis.spectrum.compare.get_ratio`,
    :func:`~radis.spectrum.compare.get_distance`,
    :func:`~radis.spectrum.compare.plot_diff`,
    :func:`~radis.spectrum.compare.get_residual_integral`,
    :meth:`~radis.spectrum.spectrum.compare_with`
    """

    if normalize:
        if isinstance(normalize, tuple):
            wrange = normalize
        else:
            wrange = ()
        var, wunit, Iunit = get_default_units(s1, s2, var=var, wunit=wunit, Iunit=Iunit)
        s1 = s1.take(var).normalize(
            wrange=wrange, normalize_how=normalize_how, wunit=wunit
        )
        s2 = s2.take(var).normalize(
            wrange=wrange, normalize_how=normalize_how, wunit=wunit
        )
        Iunit = s1.units[var]
        assert Iunit == s2.units[var]

    var, wunit, Iunit = get_default_units(s1, s2, var=var, wunit=wunit, Iunit=Iunit)
    wdiff, dI = get_diff(
        s1,
        s2,
        var,
        resample=True,
        diff_window=diff_window,
        wunit=wunit,
        Iunit=Iunit,
    )

    if ignore_nan:
        b = np.isnan(dI)
        wdiff, dI = wdiff[~b], dI[~b]

    if norm == "L2":
        output = np.sqrt((dI**2).sum()) / len(dI)
    elif norm == "L1":
        output = (np.abs(dI)).sum() / len(dI)
    else:
        raise ValueError("unexpected value for norm")

    if np.isnan(output):
        warn(
            'NaN output in residual. You should use "ignore_nan=True". Read the help of get_residual.',
            UserWarning,
        )

    return output


def get_residual_integral(
    s1: Spectrum,
    s2: Spectrum,
    var,
    ignore_nan=False,
    wunit="default",
    Iunit="default",
) -> float:
    # type: (Spectrum, Spectrum, str, bool) -> float
    """Returns integral of the difference between two spectra s1 and s2,
    relatively to the integral of spectrum s1.

    Compared to :func:`~radis.spectrum.compare.get_residual`, this tends to
    cancel the effect of the gaussian noise of an experimental spectrum.

    ::

        res = trapz(I2-I1, w1) / trapz(I1, w1)

    Note: when the considered variable is ``transmittance`` or ``transmittance_noslit``,
    the *upper* integral is used (up to 1) to normalize the integral difference

    ::

        res = trapz(I2-I1, w1) / trapz(1-I1, w1)

    Parameters
    ----------
    s1, s2: Spectrum objects
        :py:class:`~radis.spectrum.spectrum.Spectrum`
    var: str
        spectral quantity

    Other Parameters
    ----------------
    ignore_nan: boolean
        if ``True``, ignore nan in the difference between s1 and s2 (ex: out of bound)
        when calculating residual. Default ``False``. Note: ``get_residual_integral``
        will still fail if there are nan in initial Spectrum.
    wunit: str
        If ``'default'``, use first spectrum unit.
    Iunit: str
        if ``'default'``, use first spectrum unit


    Notes
    -----
    For I1, I2, the values of 'var' in s1 and s2, respectively, residual
    is calculated as::

        res = trapz(I2-I1, w1) / trapz(I1, w1)

    0 values for I1 yield nans except if I2 = I1 = 0

    when s1 and s2 dont have the size wavespace range, they are automatically
    resampled through get_diff on 's1' range


    See Also
    --------
    :func:`~radis.spectrum.compare.get_diff`,
    :func:`~radis.spectrum.compare.get_ratio`,
    :func:`~radis.spectrum.compare.get_distance`,
    :func:`~radis.spectrum.compare.get_residual`,
    :func:`~radis.spectrum.compare.plot_diff`,
    :meth:`~radis.spectrum.spectrum.compare_with`
    """
    var, wunit, Iunit = get_default_units(s1, s2, var=var, wunit=wunit, Iunit=Iunit)

    # Get data
    # ----
    w1, I1 = s1.get(var, wunit=wunit, Iunit=Iunit)
    wdiff, dI = get_diff(s1, s2, var, wunit=wunit, Iunit=Iunit, resample=True)

    # mask for 0
    if ignore_nan:
        b = np.isnan(dI)
        wdiff, dI = wdiff[~b], dI[~b]
        b = np.isnan(I1)
        w1, I1 = w1[~b], I1[~b]

    if var in ["transmittance", "transmittance_noslit"]:
        norm = 1 - np.trapz(I1, w1)
    else:
        norm = np.trapz(I1, w1)

    return np.abs(np.trapz(dI, wdiff)) / norm


def get_default_units(
    s1: Spectrum, s2: Spectrum, var=None, wunit="default", Iunit="default"
):
    """Get ``wunit``, Iunit for var; compatible with both spectra s1 and s2

    Parameters
    ----------
    s1, s2: Spectrum objects
    var: str, or None
        spectral quantity to plot (ex: ``'abscoeff'``). If None,
        plot the first one in the Spectrum from ``'radiance'``,
        ``'radiance_noslit'``, ``'transmittance'``, etc.
    wunit: ``'default'``, ``'nm'``, ``'cm-1'``, ``'nm_vac'``
        wavespace unit:  wavelength air, wavenumber, wavelength vacuum.
        If ``'default'``, use first spectrum wunit
    Iunit: str
        if ``'default'``, use first spectrum unit

    Returns
    -------
    var: str
        variable (evaluated if input was ``None``)
    wunit, Iunit: str
        units (evaluated if input was ``'default'``)
    """

    # Get defaults
    # ---
    if var is None:  # if nothing is defined, try these first:
        params = s1.get_vars()
        if "radiance" in params:
            var = "radiance"
        elif "radiance_noslit" in params:
            var = "radiance_noslit"
        elif "transmittance" in params:
            var = "transmittance"
        elif "transmittance_noslit" in params:
            var = "transmittance_noslit"
        else:
            # or plot the first variable we find
            var = list(params)[0]
            if var.replace("_noslit", "") in params:
                var = var.replace("_noslit", "")
    # ... create variable if it doesnt exist
    if var not in s1.get_vars():
        s1.get(var)
    if var not in s2.get_vars():
        s2.get(var)
    if wunit == "default":
        wunit = s1.get_waveunit()
    wunit = cast_waveunit(wunit)
    if Iunit == "default":
        try:
            s1.units[var]
        except KeyError:  # unit not defined in dictionary
            raise KeyError(
                "Iunit not defined in spectrum for variable {0}. ".format(var)
                + "Cant use default unit. Specify unit in s.units['{0}'].".format(var)
            )
        # get default (note that default "Iunit" may depend on waveunit used. Let s.get() handle this)
        _, _, _, Iunit = s1.get(
            var, wunit=wunit, Iunit="default", copy=False, return_units="as_str"
        )

    return var, wunit, Iunit


def _get_wdiff_Idiff(
    s1: Spectrum,
    s2: Spectrum,
    var=None,
    wunit="default",
    Iunit="default",
    resample=True,
    method="diff",
    diff_window=0,
    normalize=False,
    verbose=False,
):
    """Get difference/ratio between two spectra.

    Properly make sure all units are the same.

    Calls :py:func:`radis.spectrum.compare.get_diff`,  :py:func:`radis.spectrum.compare.get_ratio`
    or  :py:func:`radis.spectrum.compare.get_distance`

    Called by :py:func:`radis.spectrum.compare.plot_diff`

    Parameters
    ----------
    all:
        see :py:func:`radis.spectrum.compare.plot_diff`

    Returns
    -------
    wdiff, Iunit: list of arrays
        difference or ratio (depending on ``method``)
    methods: list of str
        methods
    """

    # Get all comparison methods:
    # ------------------------
    if isinstance(method, list):
        methods = method
    else:
        methods = [method]

    for method in methods:
        if diff_window != 0 and method != "diff":
            raise NotImplementedError("diff_window with method {0}".format(method))

    # Get data
    # ----
    if normalize:
        # copy before modifying directly in spectrum
        s1 = s1.copy(copy_lines=False)
        s2 = s2.copy(copy_lines=False)
        w1, I1 = s1.get(var, wunit=wunit, copy=False)
        w2, I2 = s2.get(var, wunit=wunit, copy=False)
        ratio = np.nanmax(I1) / np.nanmax(I2)
        I1 /= np.nanmax(I1)
        I2 /= np.nanmax(I2)

        if verbose:
            print(("Rescale factor: " + str(ratio)))

    wdiffs, Idiffs = [], []
    for method in methods:
        if not normalize:
            if method == "distance":
                wdiff, Idiff = get_distance(
                    s1, s2, var=var, wunit=wunit, Iunit=Iunit, resample=resample
                )
            elif method == "diff":
                wdiff, Idiff = get_diff(
                    s1,
                    s2,
                    var=var,
                    wunit=wunit,
                    Iunit=Iunit,
                    diff_window=diff_window,
                    resample=resample,
                )
            elif method == "ratio":
                wdiff, Idiff = get_ratio(
                    s1, s2, var=var, wunit=wunit, Iunit=Iunit, resample=resample
                )
            else:
                raise ValueError("Unknown comparison method: {0}".format(method))
            wdiffs.append(wdiff)
            Idiffs.append(Idiff)
        else:
            if method == "distance":
                raise NotImplementedError(
                    "{0} was not implemented yet for normalized spectra".format(method)
                )
            elif method == "diff":
                if not resample:
                    raise NotImplementedError()
                wdiff, Idiff = curve_substract(w1, I1, w2, I2)
            elif method == "ratio":
                wdiff, Idiff = get_ratio(
                    s1, s2, var=var, wunit=wunit, Iunit=Iunit, resample=resample
                )
            else:
                raise ValueError("Unknown comparison method: {0}".format(method))
            wdiffs.append(wdiff)
            Idiffs.append(Idiff)
    return s1, s2, wdiffs, Idiffs, methods


def plot_diff(
    s1: Spectrum,
    s2: Spectrum,
    var=None,
    wunit="default",
    Iunit="default",
    resample=True,
    method="diff",
    diff_window=0,
    show_points=False,
    label1=None,
    label2=None,
    figsize=None,
    title=None,
    nfig=None,
    normalize=False,
    yscale="linear",
    verbose=True,
    save=False,
    show=True,
    show_residual=False,
    lw_multiplier=1,
    diff_scale_multiplier=1,
    discard_centile=0,
    plot_medium="vacuum_only",
    legendargs={"loc": "best"},
    show_ruler=False,
):
    """Plot two spectra, and the difference between them. ``method=`` allows
    you to plot the absolute difference, ratio, or both.

    If waveranges dont match, ``s2`` is interpolated over ``s1``.


    Parameters
    ----------
    s1, s2: Spectrum objects
    var: str, or None
        spectral quantity to plot (ex: ``'abscoeff'``). If None,
        plot the first one in the Spectrum from ``'radiance'``,
        ``'radiance_noslit'``, ``'transmittance'``, etc.
    wunit: ``'default'``, ``'nm'``, ``'cm-1'``, ``'nm_vac'``
        wavespace unit:  wavelength air, wavenumber, wavelength vacuum.
        If ``'default'``, use first spectrum wunit
    Iunit: str
        if ``'default'``, use first spectrum unit
    method: ``'distance'``, ``'diff'``, ``'ratio'``, or list of them.
        If ``'diff'``, plot difference of the two spectra.
        If ``'distance'``, plot Euclidian distance (note that units are meaningless then)
        If ``'ratio'``, plot ratio of two spectra
        Default ``'diff'``.

        .. warning::
            with ``'distance'``, calculation scales as ~N^2 with N the number
            of points in a spectrum (against ~N with ``'diff'``). This can quickly
            override all memory.

        Can also be a list::

            method=['diff', 'ratio']

    normalize: bool
        Normalize the spectra to be plotted

    Other Parameters
    ----------------'
        plot yscale
    yscale: 'linear', 'log
    diff_window: int
        If non 0, calculates diff by offsetting s1 by ``diff_window`` number of
        units on either side, and returns the minimum. Kinda compensates for experimental
        errors on the w axis. Default 0. (look up code to understand...)
    show_points: boolean
        if ``True``, make all points appear with 'o'
    label1, label2: str
        curve names
    figsize
        figure size
    nfig: int, str
        figure number of name
    title: str
        title
    verbose: boolean
        if ``True``, plot stuff such as rescale ratio in normalize mode. Default ``True``
    save: str
        Default is ``False``. By default won't save anything, type the path of the
        destination if you want to save it (format in the name).
    show: Bool
        Default is ``True``. Will show the plots. You should switch to ``False`` if there are more than 20 calls of this function to avoid memory load during execution.
        If ``False``, you can show the plot(s) with ``plt.show()``
    show_residual: bool
        if ``True``, calculates and shows on the graph the residual in L2 norm.
        See :func:`~radis.spectrum.compare.get_residual`. ``diff_window`` is
        used in the residual calculation too. ``normalize`` has no effect.
    diff_scale_multiplier: float
        dilate the diff plot scale. Default ``1``
    discard_centile: int
        if not ``0``, discard the firsts and lasts centile when setting the limits
        of the diff window. Example::

            discard_centile=1     #  --> discards the smallest 1% and largest 1%
            discard_centile=10    #  --> discards the smallest 10% and largest 10%

        Useful to remove spikes in a ratio, for instance.
        Note that this does not change the values of the residual. It's just
        a plot feature.
        Default ``0``

    plot_medium: bool, ``'vacuum_only'``
        if ``True`` and ``wunit`` are wavelengths, plot the propagation medium
        in the xaxis label (``[air]`` or ``[vacuum]``). If ``'vacuum_only'``,
        plot only if ``wunit=='nm_vac'``. Default ``'vacuum_only'``
        (prevents from inadvertently plotting spectra with different propagation
        medium on the same graph).
    legendargs: dict
        format arguments forwarded to the legend
    show_ruler: bool
        if `True`, add a ruler tool to the Matplotlib toolbar.

        .. warning::
            still experimental in 0.9.30 ! Try it, feedback welcome !

    Returns
    -------
    fig: figure
        fig
    [ax0, ax1]: axes
        spectra and difference axis


    Examples
    --------
    Simple use::

        from radis import plot_diff
        plot_diff(s10, s50)                # s10, s50 are two spectra

    Advanced use, plotting the total power in the label, and getting the figure
    and axes handle to edit them afterwards::

        Punit = 'mW/cm2/sr'
        fig, axes = plot_diff(s10, s50, 'radiance_noslit', figsize=(18,6),
              label1='brd 10 cm-1, P={0:.2f} {1}'.format(s10.get_power(unit=Punit),Punit),
              label2='brd 50 cm-1, P={0:.2f} {1}'.format(s50.get_power(unit=Punit),Punit)
              )
        # modify fig, axes..

    See an example in :ref:`label_spectrum_howto_compare`, which produces the output below:

    .. image:: https://radis.readthedocs.io/en/latest/_images/cdsd4000_vs_hitemp_3409K.svg

    If you wish to plot in a logscale, it can be done in the following way::

        fig, [ax0, ax1] = plot_diff(s0, s1, normalize=False, verbose=False)
        ylim0 = ax0.get_ybound()
        ax0.set_yscale("log")
        ax0.set_ybound(ylim0)

    .. minigallery:: radis.plot_diff

    See Also
    --------
    :py:func:`~radis.spectrum.compare.get_diff`,
    :py:func:`~radis.spectrum.compare.get_ratio`,
    :py:func:`~radis.spectrum.compare.get_distance`,
    :py:func:`~radis.spectrum.compare.get_residual`,
    :py:meth:`~radis.spectrum.spectrum.Spectrum.compare_with`
    """
    import matplotlib.pyplot as plt
    from matplotlib import gridspec
    from matplotlib.widgets import MultiCursor

    from radis.misc.plot import fix_style, set_style

    var, wunit, Iunit = get_default_units(s1, s2, var=var, wunit=wunit, Iunit=Iunit)

    s1, s2, wdiffs, Idiffs, methods = _get_wdiff_Idiff(
        s1,
        s2,
        var=var,
        method=method,
        wunit=wunit,
        Iunit=Iunit,
        resample=resample,
        diff_window=diff_window,
        normalize=normalize,
        verbose=verbose,
    )

    # Plot
    # ----

    # Format units
    xlabel = format_xlabel(wunit, plot_medium)

    # Init figure
    set_style()
    fig = plt.figure(num=nfig, figsize=figsize)
    gs = gridspec.GridSpec(1 + len(methods), 1, height_ratios=[3] + [1] * len(methods))
    ax0 = fig.add_subplot(gs[0])
    ax0.ticklabel_format(useOffset=False)
    ax1 = []
    for i in range(len(methods)):
        ax1i = fig.add_subplot(gs[i + 1])
        ax1i.sharex(ax0)
        ax1i.ticklabel_format(useOffset=False)
        ax1.append(ax1i)

    # Plotting style
    if show_points:
        style = "-o"
    else:
        style = "-"

    # Get labels and names
    if label1 is None:
        label1 = s1.get_name()
    if label2 is None:
        label2 = s2.get_name()
    # Max label length:
    if len(label1) > 60:
        label1 = label1[:58] + "..."
    if len(label2) > 60:
        label2 = label2[:58] + "..."

    # Plot compared spectra
    # ... note: if 'normalize', s1 & s2 have been edited by _get_wdiff_Idiff
    ax0.plot(
        *s1.get(var, wunit=wunit, Iunit=Iunit),
        style,
        color="k",
        lw=3 * lw_multiplier,
        label=label1
    )
    ax0.plot(
        *s2.get(var, wunit=wunit, Iunit=Iunit),
        style,
        color="r",
        lw=1 * lw_multiplier,
        label=label2
    )
    ax0.set_yscale(yscale)

    # cosmetic changes
    Iunit = make_up_unit(Iunit, var)

    ax0.tick_params(labelbottom=False)
    if label1 is not None or label2 is not None:
        ax0.legend(**legendargs)

    # Start to 0
    if var in ["radiance_noslit", "radiance", "abscoeff", "absorbance"]:
        ax0.set_ylim(bottom=0)

    # plot difference (sorted)
    for ax1i, wdiff, Idiff in zip(ax1, wdiffs, Idiffs):
        b = np.argsort(wdiff)
        ax1i.plot(wdiff[b], Idiff[b], style, color="k", lw=1 * lw_multiplier)

    # Write labels
    ax1[-1].set_xlabel(make_up(xlabel))
    if normalize:
        fig.text(
            0.02,
            0.5,
            ("{0} (norm.)".format(make_up(var))),
            va="center",
            rotation="vertical",
        )
    else:
        fig.text(
            0.02,
            0.5,
            ("{0} ({1})".format(make_up(var), Iunit)),
            va="center",
            rotation="vertical",
        )

    # Set limits of 'diff' window
    for i, method in enumerate(methods):
        if method == "diff":
            # symmetrize error scale:
            # auto-zoom on min, max, but discard first and last centile (case of spikes / divergences)
            Idiff = Idiffs[i]
            if discard_centile:
                Idiff_sorted = np.sort(Idiff[~np.isnan(Idiff)])
                ymax = max(
                    abs(Idiff_sorted[-int(discard_centile * len(Idiff_sorted) // 100)]),
                    abs(Idiff_sorted[int(discard_centile * len(Idiff_sorted) // 100)]),
                )
            else:
                ymax = np.nanmax(abs(Idiff))
            ax1[i].set_ylim(-ymax * diff_scale_multiplier, ymax * diff_scale_multiplier)
        elif method == "distance":
            if discard_centile:
                raise NotImplementedError(
                    "discard_centile not implemented for method=distance"
                )
            _, ymax = ax1[i].get_ylim()
            ax1[i].set_ylim(0, ymax * diff_scale_multiplier)
        elif method == "ratio":
            # auto-zoom on min, max, but discard first and last centile (case of spikes / divergences)
            Idiff = Idiffs[i]
            if discard_centile:
                Idiff_sorted = np.sort(Idiff[~np.isnan(Idiff)])
                ymin = Idiff_sorted[int(discard_centile * len(Idiff_sorted) // 100)]
                ymax = Idiff_sorted[-int(discard_centile * len(Idiff_sorted) // 100)]
            else:
                ymin = np.nanmin(Idiff)
                ymax = np.nanmax(Idiff)
            ax1[i].set_ylim(
                bottom=((ymin - 1) * diff_scale_multiplier + 1),
                top=((ymax - 1) * diff_scale_multiplier + 1),
            )
        ax1[i].set_yscale(
            yscale
        ) if method != "ratio" else "linear"  # no log in 'ratio'
    #            ymax = max(abs(Idiff_sorted[len(Idiff_sorted)//100]-1),
    #                       abs(Idiff_sorted[len(-Idiff_sorted)//100]-1))
    #            ax1[i].set_ylim(ymax*diff_scale_multiplier+1, -ymax*diff_scale_multiplier+1)

    if title:
        fig.suptitle(title)
    # Fix format
    fix_style(ax=ax0)
    for ax1i in ax1:
        fix_style(ax=ax1i)
    plt.tight_layout()
    if title:
        plt.subplots_adjust(left=0.15, top=0.92)
    else:
        plt.subplots_adjust(left=0.15)

    # Plot difference text
    for i, method in enumerate(methods):
        if method == "diff":
            difftext = "diff"
            ax1[i].axhline(y=0, color="grey", zorder=-1)
        elif method == "distance":
            difftext = "distance"
            ax1[i].axhline(y=0, color="grey", zorder=-1)
        elif method == "ratio":
            difftext = "ratio"
            ax1[i].axhline(y=1, color="grey", zorder=-1)

        # Show residualget_residual
        if show_residual:
            difftext += " (residual={0:.2g})".format(
                get_residual(
                    s1, s2, var=var, norm="L2", ignore_nan=True, diff_window=diff_window
                )
            )
        pos = ax1[i].get_position()
        fig.text(0.09, pos.ymax + 0.02, difftext)

    # Add tools
    # ... Add cursors
    axes = [ax0] + ax1
    fig.cursors = MultiCursor(
        fig.canvas,
        axes,
        color="r",
        lw=1 * lw_multiplier,
        alpha=0.2,
        horizOn=False,
        vertOn=True,
    )

    # ... Add
    if show_ruler:
        from radis.tools.plot_tools import add_ruler

        add_ruler(fig, wunit=wunit, Iunit=Iunit, ax=axes[0])

    if show:
        plt.show()
    if save:
        fig.savefig(save)
        if not show:
            plt.close(fig)  # to avoid memory load

    # Return graphs
    return fig, axes


def averageDistance(s1: Spectrum, s2: Spectrum, var="radiance"):
    """Return the average distance between two spectra.
    It's important to note for fitting that if averageDistance(s1, s2)==0
    then s1 = s2

    Parameters
    ----------

    s1, s2: Spectrum objects
        spectra to be compared

    var: str, optional
        spectral quantity (ex: 'radiance', 'transmittance'...)

    Returns
    -------

    distance: float
        Average distance as in the following equation:

        .. math::

            dist = \\frac{\\sqrt{\\sum_i {(s1_i-s2_i)^2}}}{N}.

    """

    warn(
        FutureWarning(
            "will be deprecated in future versions. Use get_residual(s1, s2, norm=L2) instead"
        )
    )

    distArray = get_diff(s1, s2, var)
    distArray_Y2 = distArray[1] * distArray[1]
    N = np.size(distArray[1])
    distance = np.sqrt(np.sum(distArray_Y2)) / N
    return distance


# %% Function to compare Spectra including conditions and lines


def compare_spectra(
    first: Spectrum,
    other: Spectrum,
    spectra_only=False,
    plot=True,
    wunit="default",
    verbose=True,
    rtol=1e-5,
    ignore_nan=False,
    ignore_outliers=False,
    ignore_conditions=["calculation_time"],
    normalize=False,
    **kwargs
) -> bool:
    """Compare Spectrum with another Spectrum object.

    Parameters
    ----------
    first: type Spectrum
        a Spectrum to be compared
    other: type Spectrum
        another Spectrum to compare with
    spectra_only: boolean, or str
        if ``True``, only compares spectral quantities (in the same waveunit)
        and not lines or conditions. If str, compare a particular quantity
        name. If False, compare everything (including lines and conditions
        and populations). Default ``False``
    plot: boolean
        if ``True``, use plot_diff to plot all quantities for the 2 spectra
        and the difference between them. Default ``True``.
    wunit: 'nm', 'cm-1', 'default'
        in which wavespace to compare (and plot). If default, natural wavespace
        of first Spectrum is taken
    rtol: float
        relative difference to use for spectral quantities comparison.
        absolute difference ``atol`` is set to 0.
    ignore_nan: boolean
        if ``True``, nans are ignored when comparing spectral quantities
    ignore_outliers: boolean, or float
        if not False, outliers are discarded. i.e, output is determined by::

            out = (~np.isclose(I, Ie, rtol=rtol, atol=0)).sum()/len(I) < ignore_outliers
    ignore_conditions: list
        do not compare the metadata of these keys
    normalize: bool
        Normalize the spectra to be plotted

    Other Parameters
    ----------------
    kwargs: dict
        arguments are forwarded to :func:`~radis.spectrum.compare.plot_diff`

    Returns
    -------
    equals: boolean
        return True if spectra are equal (respective to tolerance defined by
        rtol and other input conditions)


    Examples
    --------
    Compare two Spectrum objects, or specifically the transmittance::

        s1.compare_with(s2)
        s1.compare_with(s2, 'transmittance')


    Note that you can also simply use `s1 == s2`, that uses
    :meth:`~radis.spectrum.spectrum.Spectrum.compare_with` internally::

        s1 == s2       # will return True or False
    """

    # Check inputs
    if not 0 <= ignore_outliers < 1:
        raise ValueError("ignore_outliers should be < 1, or False")
    if not isinstance(other, Spectrum):
        raise TypeError(
            "2nd object is not a Spectrum: got class {0}".format(other.__class__)
        )
    if isinstance(spectra_only, str):  # case where we compare all quantities
        if not spectra_only in first.get_vars():
            raise ValueError(
                "{0} is not a spectral quantity in our Spectrum ({1})".format(
                    spectra_only, first.get_vars()
                )
            )
        if not spectra_only in other.get_vars():
            raise ValueError(
                "{0} is not a spectral quantity in the other Spectrum ({1})".format(
                    spectra_only, other.get_vars()
                )
            )
    if verbose:  # print conditions
        what = spectra_only if isinstance(spectra_only, str) else "all quantities"
        msg = "compare {0} with rtol={1}".format(what, rtol)
        if ignore_nan:
            msg += ", ignore_nan"
        if ignore_outliers:
            msg += ", ignore_outliers={0}".format(ignore_outliers)
        print(msg)
    if not plot and len(kwargs) > 0:
        raise ValueError("Unexpected argument: {0}".format(kwargs))

    if wunit == "default":
        wunit = first.get_waveunit()

    def _compare_dataframes(df1, df2, name):
        """

        Parameters
        ----------

        df1, df2: pandas Dataframe
            lines, or vib/rovib levels dataframes

        name: str
            for error message
        """

        from pandas.testing import assert_frame_equal

        try:
            assert_frame_equal(
                df1.sort_index(axis=0).sort_index(axis=1),
                df2.sort_index(axis=0).sort_index(axis=1),
                check_names=True,
                check_column_type=False,  # solves problem in Python 2/3 dataframes (unicode/str)
            )
            out = True

        except AssertionError as err:
            if verbose:
                print("Comparing ", name)
                print(err.args[0])
            out = False

        return out

    def _compare_arrays(I, Ie):
        """Compare spectral quantities I and Ie."""

        if ignore_nan:
            b = ~(np.isnan(I) + np.isnan(Ie))
            I = I[b]
            Ie = Ie[b]

        # in all cases, we ignore if there are nans at the same
        # on both arrays (happens with convolved values in 'valid' mode)
        equal_nan = True

        if ignore_outliers:
            out = (
                ~np.isclose(I, Ie, rtol=rtol, atol=0, equal_nan=equal_nan)
            ).sum() / len(I) < ignore_outliers
        else:
            out = np.allclose(I, Ie, rtol=rtol, atol=0, equal_nan=equal_nan)

        return bool(out)

    def _display_difference(q, q0):
        error = np.nanmax(abs(q / q0 - 1))
        avgerr = np.nanmean(abs(q / q0 - 1))
        print(
            "...",
            k,
            "don't match (up to {0:.3}% diff.,".format(error * 100)
            + " average {0:.3f}%)".format(avgerr * 100),
        )

    b = True
    if isinstance(spectra_only, str):  # compare this quantity
        vars = [spectra_only]
    else:  # compare all quantities
        b = set(first.get_vars()) == set(other.get_vars())
        if not b and verbose:
            print(
                "... list of quantities dont match: {0} vs {1}".format(
                    first.get_vars(), other.get_vars()
                )
            )
        vars = [k for k in first.get_vars() if k in other.get_vars()]

    if spectra_only:
        # Compare spectral arrays
        # -----------
        for k in vars:
            w, q = first.get(k, wunit=wunit, Iunit=first.units[k])
            w0, q0 = other.get(k, wunit=wunit, Iunit=first.units[k])
            if len(w) != len(w0):
                print(
                    "Wavespaces have different length (for {0}: {1} vs {2})".format(
                        k, len(w), len(w0)
                    )
                )
                print("We interpolate one spectrum on the other one")

                from scipy.interpolate import interp1d

                if len(q) > len(q0):
                    f = interp1d(w, q, kind="cubic")
                    new_q = f(w0)
                    b1 = _compare_arrays(new_q, q0)
                    if not b1 and verbose:
                        _display_difference(new_q, q0)
                else:
                    f = interp1d(w0, q0, kind="cubic")
                    new_q0 = f(w)
                    b1 = _compare_arrays(q, new_q0)
                    if not b1 and verbose:
                        _display_difference(q, new_q0)

            else:  # no need to interpolate
                b1 = np.allclose(w, w0, rtol=rtol, atol=0)
                b1 *= _compare_arrays(q, q0)

                if not b1 and verbose:
                    _display_difference(q, q0)
            b *= b1

            if plot:
                try:
                    plot_diff(
                        first,
                        other,
                        var=k,
                        wunit=wunit,
                        normalize=normalize,
                        verbose=verbose,
                        **kwargs
                    )
                except:
                    print("... couldn't plot {0}".format(k))

    else:
        # Compare spectral variables
        # -----------
        for k in vars:
            w, q = first.get(k, wunit=wunit, Iunit=first.units[k])
            w0, q0 = other.get(k, wunit=wunit, Iunit=first.units[k])
            if len(w) != len(w0):
                print(
                    "Wavespaces have different length (for {0}: {1} vs {2})".format(
                        k, len(w), len(w0)
                    )
                )
                b1 = False
            else:
                b1 = np.allclose(w, w0, rtol=rtol, atol=0)
                b1 *= _compare_arrays(q, q0)
                if not b1 and verbose:
                    error = np.nanmax(abs(q / q0 - 1))
                    avgerr = np.nanmean(abs(q / q0 - 1))
                    print(
                        "...",
                        k,
                        "don't match (up to {0:.3}% diff.,".format(error * 100)
                        + " average {0:.3f}%)".format(avgerr * 100),
                    )
            b *= b1

            if plot:
                try:
                    plot_diff(
                        first,
                        other,
                        var=k,
                        wunit=wunit,
                        normalize=normalize,
                        verbose=verbose,
                        **kwargs
                    )
                except:
                    print("... there was an error while plotting {0}".format(k))

        # Compare conditions and units
        # -----------
        verbose_dict = "if_different" if verbose else False
        b1 = (
            compare_dict(
                first.conditions,
                other.conditions,
                verbose=verbose_dict,
                ignore_keys=ignore_conditions,
            )
            == 1
        )
        b2 = compare_dict(first.cond_units, other.cond_units, verbose=verbose_dict) == 1
        b3 = compare_dict(first.units, other.units, verbose=verbose_dict) == 1
        if not b1 and verbose:
            print("... conditions don't match")
        if not b2 and verbose:
            print("... conditions units don't match")
        if not b3 and verbose:
            print("... units don't match")
        b *= b1 * b2 * b3

        # Compare lines
        # -----------
        if first.lines is None and other.lines is None:
            b4 = True
        elif first.lines is None:
            b4 = False
        elif other.lines is None:
            b4 = False
        else:
            b4 = _compare_dataframes(first.lines, other.lines, "lines")
        if not b4 and verbose:
            print("... lines dont match")
        b *= b4

        # Compare populations
        # -----------
        if first.populations is None and other.populations is None:
            b5 = True
        elif first.populations is None:
            b5 = False
        elif other.populations is None:
            b5 = False
        else:
            # Compare keys in populations
            b5 = True
            if (
                compare_lists(
                    first.populations, other.populations, verbose="if_different"
                )
                == 1
            ):
                # same molecules, compare isotopes
                for molecule, isotopes in first.populations.items():
                    if (
                        compare_lists(
                            isotopes,
                            other.populations[molecule],
                            verbose="if_different",
                        )
                        == 1
                    ):
                        # same isotopes, compare electronic states
                        for isotope, states in isotopes.items():
                            if (
                                compare_lists(
                                    states,
                                    other.populations[molecule][isotope],
                                    verbose="if_different",
                                )
                                == 1
                            ):
                                # same electronic states, compare levels + other information
                                for state, content in states.items():
                                    for k, v in content.items():
                                        if k in ["vib", "rovib"]:
                                            b5 *= _compare_dataframes(
                                                v,
                                                other.populations[molecule][isotope][
                                                    state
                                                ][k],
                                                "populations of {0}({1})(iso{2})".format(
                                                    molecule, state, isotope
                                                ),
                                            )
                                        else:
                                            b5 *= (
                                                v
                                                == other.populations[molecule][isotope][
                                                    state
                                                ][k]
                                            )
                            else:
                                b5 = False
                                if verbose:
                                    print(
                                        "{0}(iso{1}) states are different (see above)".format(
                                            molecule, isotope
                                        )
                                    )
                    else:
                        b5 = False
                        if verbose:
                            print(
                                "{0} isotopes are different (see above)".format(
                                    molecule
                                )
                            )
            else:
                b5 = False
                if verbose:
                    print("Molecules are different (see above)")
        if not b5 and verbose:
            print("... populations dont match (see detail above)")
        b *= b5

        # Compare slit
        # -----------

        if len(first._slit) == len(other._slit) == 0:
            # no slit anywhere
            b6 = True
        elif len(first._slit) != len(other._slit):
            b6 = False
            if verbose:
                print("A spectrum has slit function array but the other doesnt")
        else:
            ws, Is = first.get_slit()
            ws0, Is0 = other.get_slit()
            if len(ws) != len(ws0):
                if verbose:
                    print("Slits have different length")
                b6 = False
            else:
                b6 = np.allclose(ws, ws0, rtol=rtol, atol=0)
                b6 *= _compare_arrays(Is, Is0)
                if not b6 and verbose:
                    print("Slit functions dont match")
        b *= b6

    return bool(b)


if __name__ == "__main__":

    from radis.test.spectrum.test_compare import _run_testcases

    print(("Test compare.py: ", _run_testcases(verbose=True)))



# -*- coding: utf-8 -*-
"""

@author: Erwan

-------------------------------------------------------------------------------


"""


import importlib
import inspect
import os
import sys
from fnmatch import translate
from os.path import basename, dirname, join
from re import IGNORECASE, compile


def getProjectRoot():
    """Return the full path of the project root."""

    return dirname(dirname(__file__))


def import_from_module(module, name):
    """Import object 'name' from module 'module' raises AttributeError if name
    doesnt exist.

    Parameters
    ----------

    module, name: str
        module path, object name
    """
    impmodule = importlib.import_module(module)
    return getattr(impmodule, name)


class Chdir:
    """because we need to change directory to get into the RADIS folder to find
    the Git files and version, when imported from another program. This class
    then ensures we get back in the correct directory.

    Examples
    --------

    Do::

        cd = Chdir(os.path.dirname(__file__))
        try:
            (...)
        except:
            (...)
        finally:
            (...)
            cd.__del__()
    """

    def __init__(self, newPath):
        self.savedPath = os.getcwd()
        os.chdir(newPath)

    def __del__(self):
        os.chdir(self.savedPath)


# %%
# ==============================================================================
# Function arguments
# ==============================================================================


class Default:
    """Contains a value. Used to know whether a function argument equal to its
    default value was explicitly given by the user or not. This allows to
    prevent user errors.

    Examples
    --------

    Check if a value is Default::

        from radis.misc.utils import Default
        a = Default("42")
        isinstance(a, Default)
        >>> True

        a.value
        >>> 42
    """

    def __repr__(self):
        return "{}".format(self.value)

    def __init__(self, value):
        self.value = value


def getarglist(function):
    """Get list of arguments in a function.

    See https://stackoverflow.com/a/41188411/5622825
    """

    if sys.version_info[0] == 2:
        from inspect import getargspec

        return getargspec(function).args

    else:
        from inspect import signature

        return list(signature(function).parameters)


def get_default_arg(func, arg):
    """Get default value of argument ``arg`` in function ``func``

    Adapted from https://stackoverflow.com/questions/12627118/get-a-function-arguments-default-value
    """
    signature = inspect.signature(func)
    items = dict(signature.parameters.items())
    if not arg in items:
        raise ValueError("Function {0} has no argument `{1}`".format(func, arg))
    elif items[arg].default is inspect.Parameter.empty:
        raise ValueError("No default value for argument `{0}` in {1}".format(arg, func))
    else:
        return items[arg].default


# %% Other stuff


class DatabankNotFound(FileNotFoundError):
    """Used when a line database is not found in radis.rc."""

    pass


# %%
# ==============================================================================
# Optional packages
# ==============================================================================


class NotInstalled(object):
    """A class to deal with optional packages. Will raise an error only if the
    package is used (but not if imported only)

    Examples
    --------

    ::

        some function = NotInstalled("you should install C drivers")
        a = some_function   # no error
        a()             # will raise the NotInstalled error and display the message

    """

    def __init__(self, name, info=""):
        self.__name = name
        self.__info = info

    def __getattr__(self, item):
        raise ImportError(
            "The {0} package is required to use this "
            "feature. {1}".format(self.__name, self.__info)
        )

    def __call__(self, *args, **kwargs):
        raise ImportError(
            "The {0} package is required to use this "
            "feature. {1}".format(self.__name, self.__info)
        )


not_installed_vaex_args = (
    "vaex",
    "You must install Vaex to use these features. Vaex is a fast, "
    + "memory-mapped DataFrame library. However is not available yet on latest Python versions. "
    + "Use Pytables (slower) as an alternative in your Radis.json config file. To use Pytables, set "
    + '"MEMORY_MAPPING_ENGINE": "pytables" and "DATAFRAME_ENGINE": "pandas"',
)
not_installed_nvidia_args = (
    "nvidia-cufft",
    "Nvidia was not installed on your computer. `nvidia-cufft` is"
    + "not available on Mac OS. ALternative solutions are being developped "
    + "with Vulkan, see https://github.com/radis/radis/pull/624. (latest update Oct. 2024). ",
)


def get_files_from_regex(path):
    """Returns a list of absolute paths of all the files whose names match the
    input regular expression."""
    directory_name = dirname(path)
    if directory_name == "":
        directory_name = "."
    regex = basename(path)

    file_names = []

    pattern = compile(translate(regex), IGNORECASE)

    for file in os.listdir(directory_name):
        if pattern.fullmatch(file):
            file_names.append(join(directory_name, file))

    return file_names


# %% Test


def _test(*args, **kwargs):

    print("Project root:", getProjectRoot())

    return True


if __name__ == "__main__":
    _test()



# -*- coding: utf-8 -*-
"""
Test names and labels
"""

import pytest

from radis.lbl.labels import (
    vib_lvl_name_cdsd_pc,
    vib_lvl_name_cdsd_pcN,
    vib_lvl_name_hitran_class1,
    vib_lvl_name_hitran_class5,
    vib_lvl_name_hitran_class5_short,
)
from radis.misc.printer import printm


@pytest.mark.fast
def test_vibrational_levels_labelling(verbose=True, *args, **kwargs):

    # If verbose: print some
    if verbose:
        printm("Some vibrational level formats:")

        printm("... CO, HITRAN format (v):\t\t", vib_lvl_name_hitran_class1(10))

        printm(
            "... CO2, HITRAN format (v1,v2,l2,v3):\t\t",
            vib_lvl_name_hitran_class5(2, 1, 1, 3),
        )
        printm(
            "... CO2, HITRAN format, short (v1v2l2v3):\t",
            vib_lvl_name_hitran_class5_short(2, 1, 1, 3),
        )

        printm("... CO2, CDSD format (p,c):\t\t", vib_lvl_name_cdsd_pc(14, 1))
        printm("... CO2, CDSD format (p,c,N):\t\t", vib_lvl_name_cdsd_pcN(14, 1, 1))

    # Do the tests

    assert vib_lvl_name_hitran_class1(10) == "(10)"

    assert vib_lvl_name_hitran_class5(2, 1, 1, 3) == "(2,1,1,3)"
    assert vib_lvl_name_hitran_class5_short(2, 1, 1, 3) == "21`1`3"

    assert vib_lvl_name_cdsd_pc(14, 1) == "(14,1)"
    assert vib_lvl_name_cdsd_pcN(14, 1, 1) == "(14,1,1)"

    return


if __name__ == "__main__":

    test_vibrational_levels_labelling()



# -*- coding: utf-8 -*-
"""
Summary
-------

Public (front-end) functions to calculate Spectrum with HITRAN / CDSD databanks.
Uses the SpectrumFactory class from `factory.py`, Spectrum from `spectrum.py`
and line survey from `line_survey.py`

Routine Listing
---------------

:func:`~radis.lbl.calc.calc_spectrum`

-------------------------------------------------------------------------------

"""


from copy import deepcopy
from os.path import exists

import numpy as np

try:  # Proper import
    from .base import get_wavenumber_range
    from .factory import SpectrumFactory
except ImportError:  # if ran from here
    from radis.lbl.factory import SpectrumFactory
    from radis.lbl.base import get_wavenumber_range

from radis import config
from radis.misc.basics import all_in
from radis.misc.utils import Default
from radis.spectrum.spectrum import Spectrum


# %%
def calc_spectrum(
    wmin=None,
    wmax=None,
    wunit=Default("cm-1"),
    Tgas=None,
    Tvib=None,
    Trot=None,
    Telec=None,
    pressure=1.01325,
    species=None,
    isotope="all",
    mole_fraction=1,
    diluent=Default(None),
    path_length=1,
    databank="hitran",
    medium="air",
    wstep=0.01,
    truncation=Default(50),
    neighbour_lines=0,
    cutoff=1e-27,
    parsum_mode="full summation",
    optimization="simple",
    chunksize=None,
    broadening_method="voigt",
    overpopulation=None,
    name=None,
    save_to="",
    use_cached=True,
    mode="cpu",
    export_lines=False,
    verbose=True,
    return_factory=False,
    **kwargs,
) -> Spectrum:
    r"""Calculate a :py:class:`~radis.spectrum.spectrum.Spectrum`.

    Can automatically download databases or use manually downloaded
    local databases, under equilibrium or non-equilibrium, with or without overpopulation,
    using either CPU or GPU.

    It is a wrapper to :py:class:`~radis.lbl.factory.SpectrumFactory` class.
    For advanced used, please refer to the aforementioned class.


    Parameters
    ----------
    wmin, wmax: float [:math:`cm^{-1}`] or `~astropy.units.quantity.Quantity`
        wavelength/wavenumber range. If no units are given, use :math:`cm^{-1}` ::

            calc_spectrum(2000, 2300, ... )   # cm-1
            calc_spectrum(4000, 4200, wunit='nm', ...)

        You can use arbitrary units::

            import astropy.units as u
            calc_spectrum(2.5*u.um, 3.0*u.um, ...)
    wunit: ``'nm'``, ``'cm-1'``
        unit for ``wmin`` and ``wmax``. Default is ``"cm-1"``.
    Tgas: float [:math:`K`]
        Gas temperature. If non equilibrium, is used for :math:`T_{translational}`.
        Default ``300`` K​
    Tvib, Trot: float [:math:`K`]
        Vibrational and rotational temperatures (for non-LTE calculations).
        If ``None``, they are at equilibrium with ``Tgas`` ​. Only applicable for molecules, not atoms.
    Telec: float [:math:`K`]
        Electronic temperature (for non-LTE calculations). If ``None``, it is at equilibrium with ``Tgas`` ​. Only implemented for atoms, not molecules.
    pressure: float [:math:`bar`] or `~astropy.units.quantity.Quantity`
        partial pressure of gas in bar. Default ``1.01325`` (1 atm)​. Use arbitrary units::

            import astropy.units as u
            calc_spectrum(..., pressure=20*u.mbar)

    species: int, str, list or ``None``
        For molecules:
            molecule id (HITRAN format) or name. For multiple molecules, use a list.
            The ``'isotope'``, ``'mole_fraction'``, ``'databank'`` and ``'overpopulation'`` parameters must then
            be dictionaries.
            If ``None``, the molecule can be inferred
            from the database files being loaded. See the list of supported molecules
            in :py:data:`~radis.db.MOLECULES_LIST_EQUILIBRIUM`
            and :py:data:`~radis.db.MOLECULES_LIST_NONEQUILIBRIUM`.
        For atoms:
            The positive or neutral atomic species. It may be given in spectroscopic notation or any form that can be converted by :py:func:`~radis.db.classes.to_conventional_name`
        Default ``None``.​
    isotope: int, list, str of the form ``'1,2'``, or ``'all'``, or dict
        isotope id

        For molecules, this is the isotopologue ID (sorted by relative density: (eg: 1: CO2-626, 2: CO2-636 for CO2) - see [HITRAN-2020]_ documentation for isotope list for all species.

        For atoms, use the isotope number of the isotope (the total number of protons and neutrons in the nucleus) - use 0 to select rows where the isotope is unspecified, in which case the standard atomic weight from the ``periodictable`` module is used when mass is required.

        If ``'all'``,
        all isotopes in database are used (this may result in larger computation
        times!).

        Default ``'all'``.

        For multiple molecules, use a dictionary with molecule names as keys ::

            isotope={'CO2':'1,2' ,  'CO':'1,2,3' }​

    mole_fraction: float or dict
        database species mole fraction. Default ``1``.

        For multiple molecules, use a dictionary with molecule names as keys ::

            mole_fraction={'CO2': 0.8, 'CO':0.2}​

    diluent: str or dictionary
            can be a string of a single diluent or a dictionary containing diluent
            name as key and its mole_fraction as value
            For single diluent ::

                diluent = 'CO2'

            For multiple diluents ::

                diluent = { 'CO2': 0.6, 'H2O':0.2}

            For free electrons, use the symbol 'e-'. Currently, only H, H2, H2, and e- are supported for atoms - any other diluents have no effect besides diluting the mole fractions of the other constituents.

            If left as ``None``, it defaults to ``'air'`` for molecules and atomic hydrogen 'H' for atoms.

    path_length: float [:math:`cm`] or `~astropy.units.quantity.Quantity`
        slab size. Default ``1`` cm​. Use arbitrary units::

            import astropy.units as u
            calc_spectrum(..., path_length=1000*u.km)

    databank: str or dict
        can be either:
        - ``'hitran'``, to fetch the latest HITRAN version
          through :py:func:`~radis.io.hitran.fetch_hitran` (download full database
          with  [HAPI]_) or :py:func:`~radis.io.query.fetch_astroquery` (download
          only the required range). To use one mode or the other, use ::

            databank=('hitran', 'full')     # download and cache full database, all isotopes
            databank=('hitran', 'range')    # download and cache required range, required isotope

        - ``'hitemp'``, to fetch the latest HITEMP version
          through :py:func:`~radis.io.hitemp.fetch_hitemp`. Downloads all lines
          and all isotopes.
        - ``'exomol'``, to fetch the latest ExoMol database
          through :py:func:`~radis.io.exomol.fetch_exomol`. To download a specific
          database use (more info in fetch_exomol) ::

            databank=('exomol', 'EBJT')   # 'EBJT' is a specific ExoMol database name

        - ``'geisa'``, to fetch the GEISA 2020 database
          through :py:func:`~radis.io.geisa.fetch_geisa`. Downloads all lines
          and all isotopes.
        - the name of a a valid database file, in which case the format is inferred.
          For instance, ``'.par'`` is recognized as ``hitran/hitemp`` format.
          Accepts wildcards ``'*'`` to select multiple files ::

            databank='PATH/TO/co_*.par'

        - ``'kurucz'`` to fetch the Kurucz linelists for atoms through :py:func:`~radis.io.kurucz.fetch_kurucz`. Downloads al lines and all isotopes.

        - the name of a spectral database registered in your ``~/radis.json``
          :ref:`configuration file <label_lbl_config_file>` ::

            databank='MY_SPECTRAL_DATABASE'

        Default ``'hitran'``. See :class:`~radis.lbl.loader.DatabankLoader` for more
        information on line databases, and :data:`~radis.misc.config.DBFORMAT` for
        your ``~/radis.json`` file format.

        For multiple molecules, use a dictionary with molecule names as keys::

            databank='hitran'     # automatic download (or 'hitemp')
            databank='PATH/TO/05_HITEMP2019.par'    # path to a file
            databank='*CO2*.par' #to get all the files that have CO2 in their names (case insensitive)
            databank='HITEMP-2019-CO'   # user-defined database in Configuration file
            databank = {'CO2' : 'PATH/TO/05_HITEMP2019.par', 'CO' : 'hitran'}  # for multiple molecules

    Other Parameters
    ----------------
    medium: ``'air'``, ``'vacuum'``
        propagating medium when giving inputs with ``'wavenum_min'``, ``'wavenum_max'``.
        Does not change anything when giving inputs in wavenumber. Default ``'air'``​
    wstep: float (:math:`cm^{-1}`)  or `'auto'`
        Resolution of wavenumber grid. Default ``0.01`` cm-1.
        If `'auto'`, it is ensured that there
        are slightly more or less than :py:data:`radis.config` ``['GRIDPOINTS_PER_LINEWIDTH_WARN_THRESHOLD']``
        points for each linewidth.

        .. note::
            wstep = 'auto' is optimized for performances while ensuring accuracy,
            but is still experimental in 0.9.30. Feedback welcome!
    truncation: float (:math:`cm^{-1}`)
        Half-width over which to compute the lineshape, i.e. lines are truncated
        on each side after ``truncation`` (:math:`cm^{-1}`) from the line center.
        If ``None``, use no truncation (lineshapes spread on the full spectral range).
        Default is ``50`` :math:`cm^{-1}`

        .. note::
                Large values (> ``50``) can induce a performance drop (computation of lineshape
                typically scale as :math:`~truncation ^2` ). The default ``50`` was
                chosen to maintain a good accuracy, and still exhibit the sub-Lorentzian
                behavior of most lines far (few hundreds :math:`cm^{-1}`) from the line center.
    neighbour_lines: float (:math:`cm^{-1}`)
        The calculated spectral range is increased (by ``neighbour_lines`` cm-1
        on each side) to take into account overlaps from out-of-range lines.
        Default is ``0`` :math:`cm^{-1}`.​
    cutoff: float (~ unit of Linestrength: :math:`cm^{-1}/(molec.cm^{-2})`)
        discard linestrengths that are lower that this, to reduce calculation
        times. ``1e-27`` is what is generally used to generate line databases such as
        CDSD. If ``0``, no cutoff. Default ``1e-27`` .
    parsum_mode: 'full summation', 'tabulation'
        how to compute partition functions, at nonequilibrium or when partition
        function are not already tabulated. ``'full summation'`` : sums over all
        (potentially millions) of rovibrational levels. ``'tabulation'`` :
        builds an on-the-fly tabulation of rovibrational levels (500 - 4000x faster
        and usually accurate within 0.1%). Default ``'full summation'``

        .. note::
            parsum_mode= 'tabulation'  is new in 0.9.30, and makes nonequilibrium
            calculations of small spectra extremely fast. Will become the default
            after 0.9.31.
    optimization : ``"simple"``, ``"min-RMS"``, ``None``
        If either ``"simple"`` or ``"min-RMS"`` LDM optimization for lineshape calculation is used:

        - ``"min-RMS"`` : weights optimized by analytical minimization of the RMS-error (See: [Spectral-Synthesis-Algorithm]_)
        - ``"simple"`` : weights equal to their relative position in the grid

        If using the LDM optimization, broadening method is automatically set to ``'fft'``.
        If ``None``, no lineshape interpolation is performed and the lineshape of all lines is calculated.
        Refer to [Spectral-Synthesis-Algorithm]_ for more explanation on the LDM method for lineshape interpolation.
        Default ``"simple"``.
    overpopulation: dict
        dictionary of overpopulation compared to the given vibrational temperature.
        Default ``None``. Example::

            overpopulation = {'CO2' : {'(00`0`0)->(00`0`1)': 2.5,
                                       '(00`0`1)->(00`0`2)': 1,
                                       '(01`1`0)->(01`1`1)': 1,
                                       '(01`1`1)->(01`1`2)': 1 }
                             }​
    export_lines: boolean
        if ``True``, saves details of all calculated lines in Spectrum. This is
        necessary to later use :py:meth:`~radis.spectrum.spectrum.Spectrum.line_survey`,
        but can take some space. Default ``False``.
    name: str
        name of the output Spectrum. If ``None``, a unique ID is generated.
    save_to: str
        save to a `.spec` file which contains absorption & emission features, all
        calculation parameters, and can be opened with :py:func:`~radis.tools.database.load_spec`.
        File can be reloaded and exported to text formats afterwards, see
        :py:meth:`~radis.spectrum.spectrum.Spectrum.savetxt`.
        If file already exists, replace.
    use_cached: boolean
        use cached files for line database and energy database. Default ``True``.​
    verbose: boolean, or int
        If ``False``, stays quiet. If ``True``, tells what is going on.
        If ``>=2``, gives more detailed messages (for instance, details of
        calculation times). Default ``True``.​
    mode: ``'cpu'``, ``'gpu'``, ``'emulated_gpu'``
        if set to ``'cpu'``, computes the spectra purely on the CPU. if set to ``'gpu'``,
        offloads the calculations of lineshape and broadening steps to the GPU
        making use of parallel computations to speed up the process. Default ``'cpu'``.
        Note that ``mode='gpu'`` requires CUDA compatible hardware to execute.
        For more information on how to setup your system to run GPU-accelerated
        methods using CUDA and Cython, check `GPU Spectrum Calculation on RADIS <https://radis.readthedocs.io/en/latest/lbl/gpu.html>`__
        To try the GPU code without an actual GPU, you can use ``mode='emulated_gpu'``.
        This will run the GPU equivalent code on the CPU.
        Only ``'cpu'`` is available for atoms.
    return_factory: bool
        if ``True``, return the :py:class:`~radis.lbl.factory.SpectrumFactory` that
        computes the spectrum. Useful to access computational parameters, the line database,
        or to start batch-computations from a first spectrum calculation. Ex::

                s, sf = calc_spectrum(..., return_factory=True)
                sf.df1  # see the lines calculated
                sf.eq_spectrum(...)  #  new calculation without reloading the database
    **kwargs: other inputs forwarded to SpectrumFactory
        For instance: ``warnings``.
        See :py:class:`~radis.lbl.factory.SpectrumFactory` documentation for more
        details on input.

    Returns
    -------
    :class:`~radis.spectrum.spectrum.Spectrum`
        Output spectrum:

        - Use the :py:meth:`~radis.spectrum.spectrum.Spectrum.get` method to retrieve a
          spectral quantity (``'radiance'``, ``'radiance_noslit'``, ``'absorbance'``, etc...)
        - Or the :py:meth:`~radis.spectrum.spectrum.Spectrum.plot` method to plot it
          directly.
        - See [1]_ to get an overview of all Spectrum methods
    :py:class:`~radis.lbl.factory.SpectrumFactory`
        if using ``return_factory=True``, the Factory that generated the spectrum is returned.
        if calculating multiple molecules, a dictionary of factories is returned


    References
    ----------
    .. [1] RADIS doc: `Spectrum how to? <https://radis.readthedocs.io/en/latest/spectrum/spectrum.html#label-spectrum>`__

    ​.. [2] RADIS GPU support: `GPU Calculations on RADIS <https://radis.readthedocs.io/en/latest/lbl/gpu.html>`__
    ​
    Examples
    --------
    Calculate a CO spectrum from the HITRAN database::

        from radis import calc_spectrum
        s = calc_spectrum(1900, 2300,         # cm-1
                          molecule='CO',
                          isotope='1,2,3',
                          pressure=1.01325,   # bar
                          Tgas=1000,
                          mole_fraction=0.1,
                          databank='hitran',  # or 'hitemp'
                          diluent = "air"     # or {'CO2': 0.1, 'air':0.8}
                          )
        s.apply_slit(0.5, 'nm')
        s.plot('radiance')

    This example uses the :py:meth:`~radis.spectrum.spectrum.Spectrum.apply_slit`
    and :py:meth:`~radis.spectrum.spectrum.Spectrum.plot` methods. See also
    :py:meth:`~radis.spectrum.spectrum.Spectrum.line_survey`::

        s.line_survey(overlay='radiance')

    Calculate a CO2 spectrum from the CDSD-4000 database::

        s = calc_spectrum(2200, 2400,   # cm-1
                          molecule='CO2',
                          isotope='1',
                          databank='/path/to/cdsd/databank/in/npy/format/',
                          pressure=0.1,  # bar
                          Tgas=1000,
                          mole_fraction=0.1,
                          mode='gpu'
                          )

        s.plot('absorbance')

    This example uses the :py:meth:`~radis.lbl.factory.SpectrumFactory.eq_spectrum_gpu` method to calculate
    the spectrum on the GPU. The databank points to the CDSD-4000 databank that has been
    pre-processed and stored in ``numpy.npy`` format.
    ​
    Refer to the online :ref:`Examples <label_examples>` for more cases, and to
    the :ref:`Spectrum page <label_spectrum>` for details on post-processing methods.

    For more details on how to use the GPU method and process the database, refer to the examples
    linked above and the documentation on :ref:`GPU support for RADIS <label_radis_gpu>`.
    ​
    Other Examples
    --------------

    .. minigallery:: radis.calc_spectrum

    References
    ----------
    **cite**: RADIS is built on the shoulders of many state-of-the-art packages and databases. If using RADIS
    to compute spectra, make sure you cite all of them, for proper reproducibility and acknowledgement of
    the work ! See :ref:`How to cite? <label_cite>` and the :py:meth:`~radis.spectrum.spectrum.Spectrum.cite`
    method.

    See Also
    --------
    :py:class:`~radis.lbl.factory.SpectrumFactory`
    """

    # Check inputs
    if "molecule" in kwargs:
        if species is not None:
            if species != kwargs["molecule"]:
                raise Exception(
                    "Both `molecule` and `species` arguments have been given and aren't equal, but `molecule` is just an alias for `species`."
                )
        species = molecule = kwargs.pop("molecule")  # remove molecule from kwargs
        # warn(
        #     DeprecationWarning(
        #         "`molecule` is deprected - use `species` instead"
        #     )
        # )
    else:
        molecule = species

    # Get wavenumber, based on whatever was given as input.
    wavenum_min, wavenum_max, input_wunit = get_wavenumber_range(
        wmin,
        wmax,
        wunit,
        kwargs.pop("wavenum_min") if "wavenum_min" in kwargs else None,
        kwargs.pop("wavenum_max") if "wavenum_max" in kwargs else None,
        kwargs.pop("wavelength_min") if "wavelength_min" in kwargs else None,
        kwargs.pop("wavelength_max") if "wavelength_max" in kwargs else None,
        medium,
        return_input_wunit=True,
    )
    # Deal with Multi-molecule mode:

    from radis.los.slabs import MergeSlabs

    if molecule is not None and type(molecule) != list:
        molecule = [molecule]  # fall back to the other case: multiple molecules

    # Stage 1. Find all molecules, whatever the user input configuration

    # ... Input arguments that CAN be dictionaries of molecules.
    DICT_INPUT_ARGUMENTS = {
        "isotope": isotope,
        "mole_fraction": mole_fraction,
        "databank": databank,
    }
    # Same, but when the values of the arguments themselves are already a dict.
    # (dealt with separately because we cannot use them to guess what are the input molecules)
    DICT_INPUT_DICT_ARGUMENTS = {"overpopulation": overpopulation}

    def _check_molecules_are_consistent(
        molecule_reference_set, reference_name, new_argument, new_argument_name
    ):
        r"""Will test that molecules set are the same in molecule_reference_set
        and new_argument, if new_argument is a dict. molecule_reference_set is
        a set of molecules (yeah!). reference_name is the name of the argument
        from which we guessed the list of molecules (used to have a clear error
        message). new_argument is the new argument to check new_argument_name
        is its name.

        Returns the set of molecules as found in new_argument, if applicable, else the molecule_reference_set (this allows us to parse all arguments too)

        Note that names are just here to provide clear error messages to the user if there is a contradiction.
        """
        if isinstance(new_argument, dict):
            if molecule_reference_set is None:  # input molecules are still unknown
                return set(new_argument.keys()), new_argument_name
            elif set(new_argument.keys()) != set(molecule_reference_set):
                raise ValueError(
                    "Keys of molecules in the {0} dictionary must be the same as given in `{1}=`, i.e: {2}. Instead, we got {3}".format(
                        new_argument_name,
                        reference_name,
                        molecule_reference_set,
                        set(new_argument.keys()),
                    )
                )
            else:
                return (
                    set(new_argument.keys()),
                    new_argument_name,
                )  # so now we changed the reference
        else:
            return molecule_reference_set, reference_name

    # Parse all inputs:
    molecule_reference_set = molecule
    reference_name = "molecule"
    for argument_name, argument in DICT_INPUT_ARGUMENTS.items():
        molecule_reference_set, reference_name = _check_molecules_are_consistent(
            molecule_reference_set, reference_name, argument, argument_name
        )

    # ... Now we are sure there are no contradictions. Just ensure we have molecules:
    if molecule_reference_set is None:

        raise ValueError(
            "Please enter the molecule(s) to calculate in the `molecule=` argument or as a dictionary in the following: {0}".format(
                list(DICT_INPUT_ARGUMENTS.keys())
            )
        )

    # Stage 2. Now we have the list of molecules. Let's get the input arguments for each of them.

    # ... Initialize and fill the master-dictionary
    molecule_dict = {}
    factory_dict = {}

    for molecule in molecule_reference_set:
        molecule_dict[molecule] = {}

        for argument_name, argument_dict in DICT_INPUT_ARGUMENTS.items():
            if isinstance(argument_dict, dict):
                # Choose the correspond value
                molecule_dict[molecule][argument_name] = argument_dict[molecule]
            else:  # argument_name is not a dictionary.
                # Let's distribute the same value to every molecule:
                molecule_dict[molecule][argument_name] = argument_dict
                # If wrong argument, it will be caught in _calc_spectrum() later.

    # ... Special case of dictionary arguments. Find out if they were given as default, or per dictionary of molecules:
    is_same_for_all_molecules = dict.fromkeys(DICT_INPUT_DICT_ARGUMENTS)
    for argument_name, argument_dict in DICT_INPUT_DICT_ARGUMENTS.items():
        if not isinstance(argument_dict, dict):
            is_same_for_all_molecules[argument_name] = True
        else:
            # Argument is a dictionary. Guess if keys are molecules, or levels.
            # Ex: overpopulation dict could be {'CO2':{'(0,0,0,1)':10}} or directly {{'(0,0,0,1)':10}}
            argument_keys = set(argument_dict.keys())
            if all_in(argument_keys, molecule_reference_set):
                is_same_for_all_molecules[argument_name] = False
            else:
                is_same_for_all_molecules[argument_name] = True
    # ... now fill them in:
    for argument_name, argument_dict in DICT_INPUT_DICT_ARGUMENTS.items():
        if is_same_for_all_molecules[argument_name]:
            for mol in molecule_reference_set:
                # copy the value for everyone
                molecule_dict[mol][argument_name] = deepcopy(
                    argument_dict
                )  # in case it gets edited.
        else:  # argument_dict keys are the molecules:
            for mol in molecule_reference_set:
                molecule_dict[mol][argument_name] = argument_dict[mol]

    # Stage 3: Now let's calculate all the spectra
    s_list = []

    """If we are computing spectrums for multiple molecules with wstep='auto',
    each spectrum can have different wstep values, thus will require resample="intersect"
    argument in MergeSlab() function to interpolate the different wstep values."""
    condition_multiple_wstep = len(molecule_dict) > 1 and wstep == "auto"
    if condition_multiple_wstep:
        wstep = [
            "auto",
            float("inf"),
        ]  # Using a list to store minimum wstep value at 1st index

    for molecule, dict_arguments in molecule_dict.items():
        kwargs_molecule = deepcopy(
            kwargs
        )  # these are the default supplementary arguments. Deepcopy ensures that they remain the same for all molecules, even if modified in _calc_spectrum

        # We add all of the DICT_INPUT_ARGUMENTS values:
        kwargs_molecule.update(**dict_arguments)

        # getting diluents for this molecule
        if len(list(molecule_dict.keys())) > 1:
            diluent_for_this_molecule = diluents_for_molecule(
                mole_fraction, diluent, molecule
            )
        elif isinstance(diluent, Default):
            diluent_for_this_molecule = diluent
        else:
            diluent_for_this_molecule = diluents_for_molecule(
                mole_fraction, diluent, molecule
            )

        generated_spectrum = _calc_spectrum_one_molecule(
            wavenum_min=wavenum_min,
            wavenum_max=wavenum_max,
            input_wunit=input_wunit,
            Tgas=Tgas,
            Tvib=Tvib,
            Trot=Trot,
            Telec=Telec,
            pressure=pressure,
            # overpopulation=overpopulation,  # now in dict_arguments
            molecule=molecule,
            # isotope=isotope,                # now in dict_arguments
            # mole_fraction=mole_fraction,    # now in dict_arguments
            path_length=path_length,
            # databank=databank,              # now in dict_arguments
            medium=medium,
            wstep=wstep,
            truncation=truncation,
            neighbour_lines=neighbour_lines,
            cutoff=cutoff,
            parsum_mode=parsum_mode,
            optimization=optimization,
            chunksize=chunksize,
            broadening_method=broadening_method,
            name=name,
            use_cached=use_cached,
            verbose=verbose,
            mode=mode,
            export_lines=export_lines,
            return_factory=return_factory,
            diluent=diluent_for_this_molecule,
            **kwargs_molecule,
        )

        if return_factory:
            factory_dict[molecule] = generated_spectrum[1]
            generated_spectrum = generated_spectrum[0]  # the spectrum
        s_list.append(generated_spectrum)

        # if condition_multiple_wstep:
        #     # Stores the minimum wstep value encountered
        #     wstep[1] = generated_spectrum.get_conditions()["wstep"]

    # Stage 4: merge all molecules and return
    if condition_multiple_wstep:
        s = MergeSlabs(*s_list, resample="intersect")
    else:
        s = MergeSlabs(*s_list)

    if save_to:
        s.store(path=save_to, if_exists_then="replace", verbose=verbose)

    if return_factory:
        if len(factory_dict) == 1:  # return unique element directly
            factory_dict = factory_dict[list(factory_dict.keys())[0]]
        return s, factory_dict
    else:
        return s


def _calc_spectrum_one_molecule(
    wavenum_min,
    wavenum_max,
    input_wunit,
    Tgas,
    Tvib,
    Trot,
    Telec,
    pressure,
    overpopulation,
    molecule,
    isotope,
    mole_fraction,
    path_length,
    databank,
    medium,
    wstep,
    truncation,
    neighbour_lines,
    cutoff,
    parsum_mode,
    optimization,
    chunksize,
    broadening_method,
    name,
    use_cached,
    verbose,
    mode,
    export_lines,
    diluent,
    return_factory=False,
    **kwargs,
) -> Spectrum:
    """See :py:func:`~radis.lbl.calc.calc_spectrum`

    Parameters
    ----------
    input_wunit: 'nm', 'nm_vac', 'cm-1'
        in which wavespace was the input given before conversion (used to keep
        default plot/get consistent with input units)
    """

    # Initialize Factory
    # ------------------

    # Check inputs

    # ... temperatures

    if Tgas is None and Trot is None:
        raise ValueError(
            "Choose either Tgas (equilibrium) or Tvib / Trot (non equilibrium)"
        )

    if Tvib is None and Trot is not None or Tvib is not None and Trot is None:
        raise ValueError("Choose both Tvib and Trot")

    # ... others
    if databank is None:
        raise ValueError("Give a databank name")

    if not "save_memory" in kwargs:
        # no need to save intermediary results as
        # factory is used once only
        kwargs["save_memory"] = True

    # if "chunksize" in kwargs:
    #     raise DeprecationWarning("use optimization= instead of chunksize=")

    def _is_at_equilibrium():
        try:
            assert Telec is None or Telec == Tgas
            assert Tvib is None or (np.array(Tvib) == np.array(Tgas)).all()
            assert Trot is None or Trot == Tgas
            assert overpopulation is None
            if "self_absorption" in kwargs:
                assert kwargs["self_absorption"]  # == True
            return True
        except AssertionError:
            return False

    _equilibrium = _is_at_equilibrium()

    # which columns to keep when loading line database
    if kwargs["save_memory"] >= 2 and _equilibrium:
        drop_columns = "all"
    else:
        drop_columns = "auto"

    # Run calculations
    sf = SpectrumFactory(
        wavenum_min=wavenum_min,
        wavenum_max=wavenum_max,
        medium=medium,
        species=molecule,
        isotope=isotope,
        pressure=pressure,
        wstep=wstep,
        truncation=truncation,
        mole_fraction=mole_fraction,
        neighbour_lines=neighbour_lines,
        cutoff=cutoff,
        parsum_mode=parsum_mode,
        verbose=verbose,
        optimization=optimization,
        chunksize=chunksize,
        broadening_method=broadening_method,
        export_lines=export_lines,
        diluent=diluent,
        **kwargs,
    )
    # Have consistent output units
    sf.input_wunit = input_wunit

    # Checking diluent other than air present
    if isinstance(diluent, str):
        if diluent == "air":
            diluent_other_than_air = False
        else:
            diluent_other_than_air = True
    else:
        diluent_other_than_air = not isinstance(diluent, Default)
        # diluent_other_than_air = len(diluent) > 1 or (
        #     len(diluent) == 1 and "air" not in diluent
        # )
    if diluent_other_than_air and databank == "exomol":
        raise NotImplementedError(
            "Only air broadening is implemented in RADIS with ExoMol. Please reach out on https://github.com/radis/radis/issues"
        )

    # Load databank
    # -------------
    sf.dataframe_type = config["DATAFRAME_ENGINE"]

    # Get databank
    if (
        databank
        in [
            "fetch",
            "hitran",
            "hitemp",
            "exomol",
            "geisa",
            "kurucz",
        ]
        or (isinstance(databank, tuple) and databank[0] == "exomol")
        or (isinstance(databank, tuple) and databank[0] == "hitran")
    ):  # mode to get databank without relying on  Line databases
        # Line database :
        if databank in ["fetch", "hitran"]:
            conditions = {
                "source": "hitran",
                "parfuncfmt": "hapi",  # use HAPI (TIPS) partition functions for equilibrium
            }
        elif isinstance(databank, tuple) and databank[0] == "hitran":
            conditions = {
                "source": "hitran",
                "database": databank[
                    1
                ],  # 'full' or 'partial', cf LoaderFactory.fetch_databank()
                "parfuncfmt": "hapi",  # use HAPI (TIPS) partition functions for equilibrium
            }
        elif databank in ["hitemp"]:
            conditions = {
                "source": "hitemp",
                "parfuncfmt": "hapi",  # use HAPI (TIPS) partition functions for equilibrium}
            }
        elif databank in ["exomol"]:
            conditions = {
                "source": "exomol",
                "parfuncfmt": "exomol",  # download & use Exo partition functions for equilibrium}
            }
        elif databank in ["geisa"]:
            conditions = {
                "source": "geisa",
                "parfuncfmt": "hapi",
                # TODO: replace with GEISA partition function someday.............
            }
        elif databank in ["kurucz"]:
            conditions = {"source": "kurucz", "parfuncfmt": "kurucz"}

        elif isinstance(databank, tuple) and databank[0] == "exomol":
            conditions = {
                "source": "exomol",
                "database": databank[1],
                "parfuncfmt": "exomol",  # download & use Exo partition functions for equilibrium}
            }
        # Partition functions :
        conditions.update(
            **{
                "levelsfmt": None,  # no need to load energies by default
                "db_use_cached": use_cached,
            }
        )
        # Rovibrational energies :
        if not _equilibrium:
            # calculate partition functions with energy levels from built-in
            # constants (not all molecules are supported!)
            conditions["levelsfmt"] = "radis"
            conditions["lvl_use_cached"] = use_cached
        # Columns to load
        if export_lines:
            conditions["load_columns"] = "all"
            conditions["extra_params"] = "all"
        elif not _equilibrium and diluent_other_than_air:
            conditions["load_columns"] = ["noneq", "diluent"]
            conditions["extra_params"] = "all"
        elif not _equilibrium:
            conditions["load_columns"] = "noneq"
        elif diluent_other_than_air:
            conditions["load_columns"] = ["equilibrium", "diluent"]
            conditions["extra_params"] = "all"
        else:
            conditions["load_columns"] = "equilibrium"

        conditions["load_energies"] = not _equilibrium
        # Details to identify lines
        conditions["parse_local_global_quanta"] = (not _equilibrium) or export_lines

        # Finally, LOAD :
        sf.fetch_databank(**conditions)
    elif exists(databank):
        conditions = {
            "path": databank,
            "drop_columns": drop_columns,
            "parfuncfmt": "hapi",  # use HAPI (TIPS) partition functions for equilibrium
            "levelsfmt": None,  # no need to load energies by default
            "db_use_cached": use_cached,
        }
        # Guess format
        if databank.endswith(".par"):
            if verbose:
                print("Inferred {0} is a HITRAN-format file.".format(databank))
            conditions["format"] = "hitran"
            # If non-equilibrium we'll also need to load the energy levels.
            if not _equilibrium:
                # calculate partition functions with energy levels from built-in
                # constants (not all molecules are supported!)
                conditions["levelsfmt"] = "radis"
                conditions["lvl_use_cached"] = use_cached
        elif databank.endswith(".h5") or databank.endswith(".hdf5"):
            if verbose:
                print(
                    "Inferred {0} is a HDF5 file with RADISDB columns format".format(
                        databank
                    )
                )
            conditions["format"] = "hdf5-radisdb"
            if not _equilibrium:
                conditions["levelsfmt"] = "radis"
        else:
            raise ValueError(
                "Couldnt infer the format of the line database file: {0}. ".format(
                    databank
                )
                + "Create a user-defined database in your ~/radis.json file "
                + "and define the format there. More information on "
                + "https://radis.readthedocs.io/en/latest/lbl/lbl.html#configuration-file"
            )

        # Columns to load
        if export_lines:
            conditions["load_columns"] = "all"
        elif not _equilibrium and diluent_other_than_air:
            conditions["load_columns"] = ["noneq", "diluent"]
        elif not _equilibrium:
            conditions["load_columns"] = "noneq"
        elif diluent_other_than_air:
            conditions["load_columns"] = ["equilibrium", "diluent"]
        else:
            conditions["load_columns"] = "equilibrium"

        conditions["load_energies"] = not _equilibrium
        # Finally, LOAD :
        sf.load_databank(**conditions)

    else:  # manual mode: get from user-defined line databases defined in ~/radis.json

        # Columns to load
        if export_lines:
            load_columns = "all"
        elif not _equilibrium and diluent_other_than_air:
            load_columns = ["noneq", "diluent"]
        elif not _equilibrium:
            load_columns = "noneq"
        elif diluent_other_than_air:
            load_columns = ["equilibrium", "diluent"]
        else:
            load_columns = "equilibrium"

        sf.load_databank(
            databank,
            load_energies=not _equilibrium,  # no need to load/calculate energies at eq.
            drop_columns=drop_columns,
            load_columns=load_columns,
        )

    #    # Get optimization strategies
    #    if lineshape_optimization == 'auto':        # NotImplemented: finally we use DLM all the time as default.
    #        if len(sf.df0) > 1e5:
    #            lineshape_optimization = 'DLM'
    #        else:
    #            lineshape_optimization = None
    #        sf.params['chunksize'] = lineshape_optimization

    if overpopulation is not None or overpopulation != {}:
        sf.misc.export_rovib_fraction = True  # required to compute Partition functions with overpopulation being taken into account

    # Calculate Spectrum
    # ------------------
    # Use the standard eq_spectrum / non_eq_spectrum functions
    if _equilibrium:
        if mode == "cpu":
            s = sf.eq_spectrum(
                Tgas=Tgas,
                mole_fraction=mole_fraction,
                path_length=path_length,
                name=name,
            )

        elif mode in ("gpu", "emulated_gpu"):
            s = sf.eq_spectrum_gpu(
                Tgas=Tgas,
                mole_fraction=mole_fraction,
                pressure=pressure,
                path_length=path_length,
                name=name,
                backend=("cpu-cuda" if mode == "emulated_gpu" else "gpu-cuda")
                # emulate=(True if mode == "emulated_gpu" else False),
            )
        else:
            raise ValueError(
                f"mode= should be one of 'cpu', 'gpu', 'emulated_gpu' (GPU code running on CPU). Got {mode}"
            )
    else:
        if mode != "cpu":
            raise NotImplementedError(mode)
        s = sf.non_eq_spectrum(
            Tvib=Tvib,
            Trot=Trot,
            Ttrans=Tgas,
            Telec=Telec,
            overpopulation=overpopulation,
            mole_fraction=mole_fraction,
            path_length=path_length,
            name=name,
        )

    if return_factory:
        return s, sf
    else:
        return s


# Function to get diluent(s) for a molecule
def diluents_for_molecule(mole_fraction, diluent, molecule):
    diluent_for_this_molecule = {}
    if isinstance(diluent, Default):
        diluent = "air"
    if isinstance(diluent, dict):
        diluent_for_this_molecule = diluent.copy()
    else:
        if isinstance(mole_fraction, dict):
            diluent_for_this_molecule[diluent] = 1 - sum(list(mole_fraction.values()))
        else:
            diluent_for_this_molecule[diluent] = 1 - mole_fraction
        # Note: will be rounded later, e.g. 1-0.7 = 0.30000000000000004 - see https://docs.python.org/3.10/tutorial/floatingpoint.html

    # Adding the other molecules from the gas mixture as diluent for the calculation of this particular molecule
    if isinstance(mole_fraction, dict):
        for other_molecule, other_fraction in mole_fraction.items():
            if other_molecule != molecule:
                if other_molecule in diluent_for_this_molecule:
                    diluent_for_this_molecule[other_molecule] += other_fraction
                else:
                    diluent_for_this_molecule[other_molecule] = other_fraction

    # finaly, remove mole fraction of diluent equal to 0 (weird input or air when molecules add up to 1)
    diluent_for_this_molecule = {
        k: v for k, v in diluent_for_this_molecule.items() if v != 0
    }
    if diluent_for_this_molecule == {}:
        return Default(None)
    else:
        return diluent_for_this_molecule


# --------------------------
if __name__ == "__main__":

    from radis.test.lbl.test_calc import _run_testcases

    print(_run_testcases(verbose=True))

# %%



# -*- coding: utf-8 -*-
"""Created on Sun Jan  3 17:52:04 2016.

@author: Erwan

Debug functions

-------------------------------------------------------------------------------
"""


# %%
# ==============================================================================
# Debug functions
# ==============================================================================


def printdbg(*args, **kwargs):
    """Function that prints only in debug mode. change this at runtime with.

    >>> radis.debug = True

    Examples
    --------

    Embed this print in a if __debug__ statement::

        if __debug__: printdbg(...)

    so that printdbg are removed by the Python preprocessor when running in
    optimize mode::

        python -O *.py

    See the ``"DEBUG_MODE"`` key in :py:attr:`radis.config`
    """

    import radis

    if radis.config["DEBUG_MODE"]:
        print("DEBUG:", *args, **kwargs)


def export(var=locals()):
    """Export local variables. Useful for debugging.

    Debugging inside a function may be tedious because you can't access the
    local variables. One of the option is to use the ipython magic::

        %debug

    Or the pdb equivalent::

        import pdb
        pdb.pm()

    Another option is to insert this export() call in the troubled function,
    before the exception occurs.


    Examples
    --------

        debug_export(locals())

    Note: you can also use  'globals().update(locals())' directly in your
    function to debug

    Note
    ----

    - seems not to work for functions nested in functions
    - 01/05 : doesn't seem to work at all.. @Erwan
    """

    globals().update(var)

    return


#
# def get_locals(func):
#    ''' Initially from :
#    http://stackoverflow.com/questions/20577806/keep-function-namespace-alive-for-debugging-in-ipython
#
#    '''
#
#    def wrap(*args, **kw):
#        sys.settrace(tracefunc)
#        try:
#            res = func(*args, **kw)
#        finally:
#            sys.settrace(None)
#        return res
#
#    def tracefunc(frame, event, arg):
#        if event == "return":
#            if frame.f_code is func.func_code:
#                wrap.last_res = frame.f_locals
#        return tracefunc
#
#    return wrap



import numpy as np
from scipy.constants import c, h, k

c_cm = 100 * c
c2 = h * c_cm / k


def init_L_params(na, gamma_arr, verbose=False):

    if verbose >= 2:
        print("Initializing Lorentzian parameters ")

    # The entire list of widths is first checked for minima, then for maxima.
    result = []
    for minmax in (np.min, np.max):

        # Remove duplicates
        na_gamma_arr = np.zeros((na.size, 2), dtype=np.float32)
        na_gamma_arr[:, 0] = na
        na_gamma_arr[:, 1] = minmax(gamma_arr, axis=0)

        unique_lines = np.unique(na_gamma_arr.reshape(2 * na.size).view(np.int64))
        unique_lines = unique_lines.view(np.float32).reshape(unique_lines.size, 2)

        # Only keep extremes
        test_dict = {}
        for (na_i, gamma_i) in unique_lines:
            try:
                test_dict[na_i] = minmax((gamma_i, test_dict[na_i]))

            except (KeyError):
                test_dict[na_i] = gamma_i

        # Check which ones are really at the top:

        keys = sorted(test_dict.keys(), reverse=(minmax == np.min))
        A = [keys[0]]
        B = [np.log(test_dict[keys[0]])]
        X = [-np.inf]

        for key in keys[1:]:
            for i in range(len(X)):
                xi = (np.log(test_dict[key]) - B[i]) / (A[i] - key)
                if xi >= X[i]:
                    if i < len(X) - 1:
                        if xi < X[i + 1]:
                            break
                    else:
                        break

            A = A[: i + 1] + [key]
            B = B[: i + 1] + [np.log(test_dict[key])]
            X = X[: i + 1] + [xi]

        X = X[1:] + [np.inf]
        result.append((A, B, X))

    param_data = tuple(result)

    if verbose >= 2:
        print("done!")

    global _L_param_data
    _L_param_data = param_data


def init_G_params(log_2vMm, verbose=False):

    if verbose >= 2:
        print("Initializing Gaussian parameters")

    param_data = (np.min(log_2vMm), np.max(log_2vMm))

    if verbose >= 2:
        print("done!")

    global _G_param_data
    _G_param_data = param_data


def init_Q(Q_intp_list):
    global _Q_intp_list
    _Q_intp_list = Q_intp_list


def set_L_params(init_h, iter_h, epsilon=1e-4):

    global _L_param_data
    result = []
    for params in _L_param_data:
        A, B, X = params
        i = 0
        while X[i] < iter_h.log_rT:
            i += 1
        result.append(iter_h.log_rT * A[i] + B[i] + iter_h.log_2p)
    log_wL_min, log_wL_max = result

    log_wL_max += epsilon
    N = int(np.ceil((log_wL_max - log_wL_min) / init_h.dxL) + 1)

    iter_h.log_wL_min = log_wL_min
    iter_h.N_L = N


def set_G_params(init_h, iter_h, epsilon=1e-4):
    global _G_param_data

    log_2vMm_min, log_2vMm_max = _G_param_data
    log_wG_min = log_2vMm_min + iter_h.hlog_T
    log_wG_max = log_2vMm_max + iter_h.hlog_T
    log_wG_max += epsilon

    N = int(np.ceil((log_wG_max - log_wG_min) / init_h.dxG) + 1)

    iter_h.log_wG_min = log_wG_min
    iter_h.N_G = N


def set_pTQ(p, T, mole_fraction, iter_h, l=1.0, slit_FWHM=0.0):
    """


    Parameters
    ----------
    p : float
        pressure [bar].
    T : float
        temperature [K].
    mole_fraction : float
    iter_h : TYPE
        DESCRIPTION.
    l : TYPE, optional
        DESCRIPTION. The default is 1.0.
    slit_FWHM : TYPE, optional
        DESCRIPTION. The default is 0.0.

    Returns
    -------
    None.

    """
    global _Q_intp_list
    iter_h.p = p  # bar
    iter_h.log_2p = np.log(2 * p)
    iter_h.hlog_T = 0.5 * np.log(T)
    iter_h.log_rT = np.log(296.0 / T)
    iter_h.c2T = -c2 / T
    iter_h.N = p * 1e5 / (1e6 * k * T)  # cm-3
    iter_h.x[0] = mole_fraction  # self
    iter_h.x[1] = 1 - mole_fraction  # air
    iter_h.l = l
    iter_h.slit_FWHM = slit_FWHM

    for i in range(len(_Q_intp_list)):
        iter_h.Q[i] = _Q_intp_list[i](T)



# -*- coding: utf-8 -*-
"""Created on Tue Jul 31 16:28:43 2018.

@author: erwan
"""


def morse_increment(Dzero, we, wexe=0, weye=0, weze=0, weae=0, webe=0, wece=0):
    """Get Morse potential energy increment correction for given molecule.

    Parameters
    ----------

    Dzero: cm-1
        dissociation energy

    we, wexe, etc.: cm-1
        spectroscopic parameters

    Notes
    -----

    Internal:

        Dzero = self.Ediss
        d_equil = (Dzero+0.5*we-0.25*wexe+0.125*weye
                   + 0.0625*weze+0.03125*weae+0.015625*webe
                   + 7.8125e-3*wece)

        v_inc = we**2 / (2*d_equil)

    References
    -----------

    Specair
    """

    d_equil = (
        Dzero
        + 0.5 * we
        - 0.25 * wexe
        + 0.125 * weye
        + 0.0625 * weze
        + 0.03125 * weae
        + 0.015625 * webe
        + 7.8125e-3 * wece
    )

    v_inc = we**2 / (2 * d_equil)

    return v_inc



# -*- coding: utf-8 -*-
"""Test radis import.

Start-up time can be monitored with Python 3.7:
https://dev.to/methane/how-to-speed-up-python-application-startup-time-nkf

See for instance:

https://stackoverflow.com/questions/16373510/improving-speed-of-python-module-import

which uses tuna

https://github.com/nschloe/tuna
"""

import radis

assert radis  # silences pyflakes



# -*- coding: utf-8 -*-
"""
Created on Tue Jan 26 22:40:51 2021

@author: erwan


https://stackoverflow.com/questions/55610891/numpy-load-from-io-bytesio-stream
https://stupidpythonideas.blogspot.com/2014/07/three-ways-to-read-files.html

"""

import re
import urllib.request
from os.path import basename, commonpath, join
from typing import Union

import numpy as np

try:
    from .dbmanager import DatabaseManager
    from .hitranapi import columns_2004, parse_global_quanta, parse_local_quanta
    from .tools import (
        _create_dtype,
        _get_linereturnformat,
        _ndarray2df,
        replace_PQR_with_m101,
    )
except ImportError:  # ran from here
    if __name__ == "__main__":  # running from this file, as a script
        from radis.api.dbmanager import DatabaseManager
        from radis.api.hitranapi import (
            columns_2004,
            parse_global_quanta,
            parse_local_quanta,
        )
        from radis.io.tools import (
            _create_dtype,
            _get_linereturnformat,
            _ndarray2df,
            replace_PQR_with_m101,
        )
    else:
        raise

from radis.db import MOLECULES_LIST_NONEQUILIBRIUM
from radis.misc.progress_bar import ProgressBar

HITEMP_MOLECULES = ["H2O", "CO2", "N2O", "CO", "CH4", "NO", "NO2", "OH"]


def keep_only_relevant(
    inputfiles, wavenum_min=None, wavenum_max=None, verbose=True
) -> Union[list, float, float]:
    """Parser file names for ``wavenum_format`` (min and max) and only keep
    relevant files if the requested range is ``[wavenum_min, wavenum_max]``

    Returns
    -------
    relevant: list of relevant files
    files_wmin, files_wmax: (float, float) : wavenum min & max of relevant range
    """
    wavenum_format = r"\d{5}"
    relevantfiles = []
    files_wmin = np.inf
    files_wmax = 0
    for filepath in inputfiles:
        file = basename(filepath)
        fname_wmin, fname_wmax = re.findall(wavenum_format, file)
        relevant = False
        if wavenum_min is not None and wavenum_max is not None:
            if (float(fname_wmax) > wavenum_min) and (float(fname_wmin) < wavenum_max):
                # strict '>' :  we exclude "CO2-02_02250-02500_HITEMP2010.h5'" if calculating 2500 - 3000 cm-1
                # strict '<' :  we exclude "CO2-02_03000-03250_HITEMP2010.h5" if calculating 2500 - 3000 cm-1
                relevant = True
        elif wavenum_min is not None:
            if float(fname_wmax) > wavenum_min:
                # strict '>' :  we exclude "CO2-02_02250-02500_HITEMP2010.h5'" if calculating 2500 - 3000 cm-1
                relevant = True
        elif wavenum_max is not None:
            if float(fname_wmin) < wavenum_max:
                # strict '<' :  we exclude "CO2-02_03000-03250_HITEMP2010.h5" if calculating 2500 - 3000 cm-1
                relevant = True
        else:
            relevant = True
        if relevant:
            relevantfiles.append(filepath)
            files_wmin = min(float(fname_wmin), files_wmin)
            files_wmax = max(float(fname_wmax), files_wmax)

    if verbose and relevantfiles != []:
        if len(relevantfiles) > 1:
            folder = commonpath(relevantfiles)
            # file_list = [x.replace(folder+'\\', '') for x in relevantfiles]
            # print(f"In {folder} keep only relevant input files: {file_list}")
            print(f"In ``{folder}`` keep only relevant input files:")
            for file in relevantfiles:
                print(file.replace(folder + "\\", ""))
        else:
            print(f"Keep only relevant input file: {relevantfiles}")

    return relevantfiles, files_wmin, files_wmax


#%%
def get_last(b):
    """Get non-empty lines of a chunk b, parsing the bytes."""
    element_length = np.vectorize(lambda x: len(x.__str__()))(b)
    non_zero = element_length > element_length[-1]
    threshold = non_zero.argmin() - 1
    assert (non_zero[: threshold + 1] == 1).all()
    assert (non_zero[threshold + 1 :] == 0).all()
    return b[non_zero]


class HITEMPDatabaseManager(DatabaseManager):
    def __init__(
        self,
        name,
        molecule,
        local_databases,
        engine="default",
        verbose=True,
        chunksize=100000,
        parallel=True,
    ):
        r"""
        See Also
        --------
        HITEMPDatabaseManager is compatible with Exojax :py:class:`exojax.spec.api.MdbHitemp`

        """
        super().__init__(
            name,
            molecule,
            local_databases,
            engine,
            verbose=verbose,
            parallel=parallel,
        )
        self.chunksize = chunksize
        self.downloadable = True
        self.base_url = None
        self.Nlines = None
        self.wmin = None  # available on HITEMP website. See HITEMPDatabaseManager.fetch_url_Nlines_wmin_wmax
        self.wmax = None  # available on HITEMP website. See HITEMPDatabaseManager.fetch_url_Nlines_wmin_wmax
        self.urlnames = None

    def fetch_url_Nlines_wmin_wmax(self, hitemp_url="https://hitran.org/hitemp/"):
        r"""requires connexion"""

        molecule = self.molecule

        if (
            self.base_url is not None
            and self.Nlines is not None
            and self.wmin is not None
            and self.wmax is not None
        ):
            return self.base_url, self.Nlines, self.wmin, self.wmax

        else:

            response = urllib.request.urlopen(hitemp_url)

            # Alternative to return a Pandas Dataframe :
            # ... Doesnt work because missing <tr> in HITEMP website table for N2O

            # soup = BeautifulSoup(
            #         response, features="lxml"
            #     )
            # table = soup.find(lambda tag: tag.name=='table' and tag.has_attr('id') and tag['id']=="hitemp-molecules-table")

            # def tableDataText(table):
            #     """Parses a html segment started with tag <table> followed
            #     by multiple <tr> (table rows) and inner <td> (table data) tags.
            #     It returns a list of rows with inner columns.
            #     Accepts only one <th> (table header/data) in the first row.

            #     From https://stackoverflow.com/a/58274853/5622825
            #     """

            #

            #     def rowgetDataText(tr, coltag='td'): # td (data) or th (header)
            #         return [td.get_text(strip=True) for td in tr.find_all(coltag)]
            #     rows = []
            #     trs = table.find_all('tr')
            #     headerrow = rowgetDataText(trs[0], 'th')
            #     if headerrow: # if there is a header row include first
            #         rows.append(headerrow)
            #         trs = trs[1:]
            #     for tr in trs: # for every table row
            #         rows.append(rowgetDataText(tr, 'td') ) # data row

            #     df = pd.DataFrame(rows[1:], columns=rows[0])
            #     df.index = df.Formula

            #     return df

            # df = tableDataText(table)

            text = response.read().decode()
            text = text[
                text.find(
                    '<table id="hitemp-molecules-table" class="selectable-table list-table">'
                ) : text.find("</table>")
            ]
            text = re.sub(r"<!--.+?-->\s*\n", "", text)  # remove commented lines
            html_molecule = re.sub(r"(\d{1})", r"(<sub>\1</sub>)", molecule)
            text = text[
                re.search(
                    "<td>(?:<strong>)?" + html_molecule + "(?:</strong>)?</td>", text
                ).start() :
            ]
            lines = text.splitlines()

            Nlines = int(re.findall(r"(\d+)", lines[3].replace("&nbsp;", ""))[0])
            wmin = int(re.findall(r"(\d+)", lines[4].replace("&nbsp;", ""))[0])
            wmax = int(re.findall(r"(\d+)", lines[5].replace("&nbsp;", ""))[0])
            url = "https://hitran.org" + re.findall(r'href="(.+?)"', lines[7])[0]

            self.base_url, self.Nlines, self.wmin, self.wmax = url, Nlines, wmin, wmax

        return url, Nlines, wmin, wmax

    def fetch_urlnames(self):
        r"""requires connection"""

        if self.urlnames is not None:
            return self.urlnames

        molecule = self.molecule

        if molecule in ["H2O", "CO2"]:

            base_url, Ntotal_lines_expected, _, _ = self.fetch_url_Nlines_wmin_wmax()
            response = urllib.request.urlopen(base_url)
            response_string = response.read().decode()
            inputfiles = re.findall('href="(\S+.zip)"', response_string)

            urlnames = [join(base_url, f) for f in inputfiles]

        elif molecule in HITEMP_MOLECULES:
            url, Ntotal_lines_expected, _, _ = self.fetch_url_Nlines_wmin_wmax()
            urlnames = [url]
        else:
            raise KeyError(
                f"Please choose one of HITEMP molecules : {HITEMP_MOLECULES}. Got '{molecule}'"
            )

        self.urlnames = urlnames

        return urlnames

    def keep_only_relevant(
        self, inputfiles, wavenum_min=None, wavenum_max=None, verbose=True
    ) -> list:
        r"""For CO2 and H2O, return only relevant files for given wavenumber range.

        If other molecule, return the file anyway.
        see :py:func:`radis.api.hitempapi.keep_only_relevant`"""
        if self.molecule in ["CO2", "H2O"]:
            inputfiles, _, _ = keep_only_relevant(
                inputfiles, wavenum_min, wavenum_max, verbose
            )
        return inputfiles

    def get_linereturn_format(self, opener, urlname, columns):

        with opener.open(urlname) as gfile:  # locally downloaded file
            dt = _create_dtype(
                columns, "a2"
            )  # 'a2' allocates space to get \n or \n\r for linereturn character
            b = np.zeros(1, dtype=dt)
            try:
                gfile.readinto(b)
            except EOFError as err:
                raise ValueError(
                    f"End of file while parsing file {opener.abspath(urlname)}. May be due to download error. Delete file ?"
                ) from err
            linereturnformat = _get_linereturnformat(b, columns)

        return linereturnformat

    def parse_to_local_file(
        self,
        opener,
        urlname,
        local_file,
        pbar_active=True,
        pbar_t0=0,
        pbar_Ntot_estimate_factor=None,
        pbar_Nlines_already=0,
        pbar_last=True,
    ):
        r"""Uncompress ``urlname`` into ``local_file``.
        Also add metadata

        Parameters
        ----------
        opener: an opener with an .open() command
        gfile : file handler. Filename: for info"""

        # Get linereturn (depends on OS, but file may also have been generated
        # on a different OS. Here we simply read the file to find out)
        columns = columns_2004
        chunksize = self.chunksize
        verbose = self.verbose
        molecule = self.molecule

        if not verbose:
            pbar_active = False

        linereturnformat = self.get_linereturn_format(opener, urlname, columns)

        Nlines = 0
        Nlines_raw = 0
        Nlines_tot = Nlines + pbar_Nlines_already
        _, Ntotal_lines_expected, _, _ = self.fetch_url_Nlines_wmin_wmax()
        if pbar_Ntot_estimate_factor:
            # multiply Ntotal_lines_expected by pbar_Ntot_estimate_factor
            # (accounts for total lines divided in number of files, and
            # not all files downloaded)
            Ntotal_lines_expected = int(
                Ntotal_lines_expected * pbar_Ntot_estimate_factor
            )
        pb = ProgressBar(N=Ntotal_lines_expected, active=pbar_active, t0=pbar_t0)
        wmin = np.inf
        wmax = 0

        writer = self.get_datafile_manager()

        with opener.open(urlname) as gfile:  # locally downloaded file

            dt = _create_dtype(columns, linereturnformat)

            if verbose:
                print(f"Download complete. Parsing {molecule} database to {local_file}")

            # assert not(exists(local_file))

            b = np.zeros(chunksize, dtype=dt)  # receives the HITRAN 160-character data.

            for nbytes in iter(lambda: gfile.readinto(b), 0):

                if not b[-1]:
                    # End of file flag within the chunk (but does not start
                    # with End of file flag) so nbytes != 0
                    b = get_last(b)

                df = _ndarray2df(b, columns, linereturnformat)

                # Post-processing :
                # ... Add local quanta attributes, based on the HITRAN group
                df = parse_local_quanta(df, molecule, verbose=verbose)

                # ... Add global quanta attributes, based on the HITRAN class
                df = parse_global_quanta(df, molecule, verbose=verbose)

                # Switch 'P', 'Q', 'R' to -1, 0, 1
                if "branch" in df:
                    replace_PQR_with_m101(df)

                writer.write(local_file, df, append=True)

                wmin = np.min((wmin, df.wav.min()))
                wmax = np.max((wmax, df.wav.max()))

                Nlines += len(df)
                Nlines_tot += len(df)
                Nlines_raw += len(b)
                if pbar_Ntot_estimate_factor is None:
                    pbar_Ntot_message = f"{Ntotal_lines_expected:,} lines"
                else:
                    pbar_Ntot_message = f"~{Ntotal_lines_expected:,} lines (estimate)"
                pb.update(
                    Nlines_tot,
                    message=f"  Parsed {Nlines_tot:,} / {pbar_Ntot_message}. Wavenumber range {wmin:.2f}-{wmax:.2f} cm-1 is complete.",
                )
                # Reinitialize for next read
                b = np.zeros(
                    chunksize, dtype=dt
                )  # receives the HITRAN 160-character data.
        writer.combine_temp_batch_files(local_file)  # used for vaex mode only
        if pbar_last:
            pb.update(
                Nlines_tot,
                message=f"  Parsed {Nlines_tot:,} / {Nlines_tot:,} lines. Wavenumber range {wmin:.2f}-{wmax:.2f} cm-1 is complete.",
            )
            pb.done()
        else:
            print("")

        # Check number of lines is consistent
        assert Nlines == Nlines_raw

        # Add metadata
        from radis import __version__

        writer.add_metadata(
            local_file,
            {
                "wavenumber_min": wmin,
                "wavenumber_max": wmax,
                "download_date": self.get_today(),
                "download_url": urlname,
                "total_lines": Nlines_raw,
                "version": __version__,
            },
        )

        return Nlines

    def register(self):
        r"""register in ~/radis.json"""

        local_files, urlnames = self.get_filenames()
        info = f"HITEMP {self.molecule} lines ({self.wmin:.1f}-{self.wmax:.1f} cm-1) with TIPS-2017 (through HAPI) for partition functions"

        if self.molecule in ["CO2", "H2O"]:
            info = "(registered files will be downloaded only when required) " + info

        dict_entries = {
            "info": info,
            "path": local_files,
            "format": "hitemp-radisdb",
            "parfuncfmt": "hapi",
            "wavenumber_min": self.wmin,
            "wavenumber_max": self.wmax,
            "download_date": self.get_today(),
            "download_url": urlnames,
        }

        # Add energy level calculation
        if self.molecule in MOLECULES_LIST_NONEQUILIBRIUM:
            dict_entries[
                "info"
            ] += " and RADIS spectroscopic constants for rovibrational energies (nonequilibrium)"
            dict_entries["levelsfmt"] = "radis"

        super().register(dict_entries)


#%%

if __name__ == "__main__":

    import pytest

    print("Testing factory:", pytest.main(["../test/io/test_hitemp.py"]))



# -*- coding: utf-8 -*-
"""
Created on Mon May  7 17:34:52 2018

@author: erwan
"""

from os.path import exists
from shutil import rmtree

import matplotlib.pyplot as plt
import numpy as np
import pytest

from radis import config
from radis.lbl import SpectrumFactory
from radis.misc.printer import printm
from radis.misc.utils import NotInstalled, not_installed_vaex_args
from radis.test.utils import getTestFile, setup_test_line_databases

try:
    import vaex
except ImportError:
    vaex = NotInstalled(*not_installed_vaex_args)

pytestmark = pytest.mark.random_order(
    disabled=True
)  # to make sure HITEMP-CO2-TEST is downloaded in 1st test and used after


@pytest.mark.needs_connection
@pytest.mark.fast
def test_retrieve_from_database(
    plot=False, verbose=True, warnings=True, *args, **kwargs
):
    """Test autoretrieve from a database:

    first generate an empty :py:class:`~radis.tools.database.SpecDatabase`
    associated to a :py:class`~radis.lbl.factory.SpectrumFactory`,
    then calculate a first spectrum, then calculate it again and make sure
    it is retrieved from the database
    """

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    temp_database_name = "temp_spec_database"

    try:
        if verbose:
            printm(">>> test_retrieve_from_database")

        assert not exists(temp_database_name)

        setup_test_line_databases()  # add HITEMP-CO2-TEST in ~/radis.json if not there

        sf = SpectrumFactory(
            2284.2,
            2284.6,
            wstep=0.001,  # cm-1
            pressure=20 * 1e-3,  # bar
            cutoff=0,
            path_length=0.1,
            mole_fraction=400e-6,
            molecule="CO2",
            isotope=[1],
            medium="vacuum",
            truncation=5,
            export_populations="rovib",
            verbose=verbose,
        )
        sf.warnings["MissingSelfBroadeningWarning"] = "ignore"
        sf.init_databank(
            "HITEMP-CO2-TEST",
            load_columns="noneq",
        )  # unlike load_databank, will automatically be built when needed
        db = sf.init_database(temp_database_name, autoretrieve=True)

        assert len(db) == 0

        # Calculate a first spectrum
        s1 = sf.non_eq_spectrum(2000, 1000)

        assert len(db) == 1
        # Note that the Spectrum in database is not s1 because populations
        # are not stored by default

        # hack: now change the `autoretrieve` status to `force` to make sure
        # the spectrum is retrieved. Else, an error will be raised
        sf.autoretrievedatabase = "force"

        import radis

        radis.debug = True
        # Calculate spectrum under the same conditions
        s2 = sf.non_eq_spectrum(2000, 1000)

        # Note that s1 == s2 won't work because populations are not stored
        # by default in the database
        assert s1.compare_with(s2, spectra_only=True, plot=plot)

    finally:
        rmtree(temp_database_name)


@pytest.mark.fast
def test_ignore_cached_files():
    """
    Previous implementation of RADIS saved the cached h5 files generated while reading the
    dataset in the same directory from where the data was being read. Using a wildcard input
    such as `path = "cdsd_hitemp_09_frag*"` in such case led to the cached files present
    in directory to also being loaded and treated as the dataset files. This resulted in
    an error due to the differences in the way data is stored in h5 files versus in dataset
    files such as par, txt, etc.

    Reference: `https://github.com/radis/radis/issues/121`
    """

    sf = SpectrumFactory(wavenum_min=2000, wavenum_max=3000, pressure=1)

    file_dir = getTestFile("cdsd_hitemp_09_fragment.txt")
    test_file = file_dir[:-8] + "*"
    sf.load_databank(path=test_file, format="cdsd-hitemp", parfuncfmt="hapi")

    try:
        sf.load_databank(path=test_file, format="cdsd-hitemp", parfuncfmt="hapi")
    except UnicodeDecodeError as err:
        raise UnicodeDecodeError(
            "Couldn't load database the 2nd time. This may be due to cache files trying to be read as normal files"
        ) from err


@pytest.mark.fast
def test_ignore_irrelevant_files(*args, **kwargs):
    """
    Implemented in https://github.com/radis/radis/pull/185

    - Test that calculating with cache files outside the spectral range they
    are properly ignored
    """

    from radis.misc.warning import EmptyDatabaseError, IrrelevantFileWarning

    # Regenerate .h5 cache file
    sf = SpectrumFactory(wavenum_min=2280, wavenum_max=2290)
    test_file = getTestFile("cdsd_hitemp_09_fragment.txt")
    sf.load_databank(
        path=test_file, format="cdsd-hitemp", parfuncfmt="hapi", db_use_cached="regen"
    )
    assert exists(test_file.replace(".txt", ".h5"))
    # Also note that there was no EmptyDatabaseError : file was properly loaded!

    # Load same .h5 cache file in another spectral range
    # ... Expect an error in use_cached = 'force' (file properly detected as
    # irrelevant)
    sf2 = SpectrumFactory(wavenum_min=100000, wavenum_max=100002)
    with pytest.raises(IrrelevantFileWarning):
        sf2.load_databank(
            path=test_file,
            format="cdsd-hitemp",
            parfuncfmt="hapi",
            db_use_cached="force",
        )

    # Again without 'force'
    # ... Expect no IrrelevantFile error, however range should be empty.
    with pytest.raises(EmptyDatabaseError):
        sf3 = SpectrumFactory(wavenum_min=100000, wavenum_max=100002)
        sf3.load_databank(path=test_file, format="cdsd-hitemp", parfuncfmt="hapi")


@pytest.mark.fast
def test_custom_abundance(verbose=True, plot=False, *args, **kwargs):
    """Test calculations with custom abundances

    set the terrestrial abundance to False to
    force computation of linestrength from Einstein coefficients instead
    of rescaling parameters

    :py:meth:`~radis.lbl.base.DatabankLoader.get_abundance`,
    :py:meth:`~radis.lbl.base.DatabankLoader.set_abundance`
    """

    sf = SpectrumFactory(
        2284.2,
        2284.6,
        wstep=0.001,  # cm-1
        pressure=20 * 1e-3,  # bar
        cutoff=0,
        path_length=0.1,
        mole_fraction=400e-6,
        molecule="CO2",
        isotope="1,2",
        medium="vacuum",
        truncation=5,
        verbose=0,
    )
    sf.warnings["MissingSelfBroadeningWarning"] = "ignore"
    sf.load_databank("HITEMP-CO2-TEST", load_columns="all")

    #%% Compute a spectrum ; set the terrestrial abundance to False to
    # force computation of linestrength from Einstein coefficients instead
    # of rescaling parameters
    s = sf.eq_spectrum(2000)
    sf.molparam.terrestrial_abundances = False
    s2 = sf.eq_spectrum(2000)
    assert (
        s != s2
    )  # compare spectra; we should see small differences due to different methods
    assert s.compare_with(s2, "abscoeff", rtol=0.5e-2, plot=False)
    # assert np.allclose(2 * s2.get("abscoeff")[1], s.get("abscoeff")[1])

    if plot:
        from radis import plot_diff

        plot_diff(s, s2)

    #%% Now restart by changing the abundance itself ; as a user would do

    # ... check abundance get/set functions work well
    abundance12_0 = sf.get_abundance("CO2", [1, 2])

    if verbose:
        print("Abundance: CO2[1]", sf.get_abundance("CO2", 1))
        print("Abundance: CO2[1,2]", sf.get_abundance("CO2", [1, 2]))

    sf.set_abundance("CO2", [1, 2], abundance12_0 / 3)
    assert all(sf.get_abundance("CO2", [1, 2]) == abundance12_0 / 3)
    sf.set_abundance("CO2", 1, abundance12_0[0] / 3)
    sf.set_abundance("CO2", 2, abundance12_0[1] / 3)
    assert all(sf.get_abundance("CO2", [1, 2]) == abundance12_0 / 3)

    # Below we compute a spectrum with the abundance divided by 3; and then
    # check that the resulting absorption coefficient is 3x smaller

    s3 = sf.eq_spectrum(2000)
    assert s.compare_with((3 * s3.take("abscoeff")), "abscoeff", rtol=0.5e-2, plot=True)


@pytest.mark.needs_connection
@pytest.mark.download_large_databases
@pytest.mark.skipif(isinstance(vaex, NotInstalled), reason="Vaex not available")
def test_vaex_and_pandas_dataframe_fetch_databank():
    """
    Compares dataframes fetched and manipulated with vaex with that of pandas.
    Every column of dataframe in vaex format is compared with corresponding column
    of dataframe in pandas format .
    Here expected
    - All columns are same for dataframe in pandas and vaex format
    - Ensure dataframes are same across all the databanks available in RADIS
    Added in https://github.com/radis/radis/pull/580
    """
    import pandas
    import vaex

    from radis.lbl import SpectrumFactory

    sf = SpectrumFactory(
        1500,
        2000,
        molecule="CO",
        isotope="1,2",  # testing only 2 isotopes to save (a bit of) time
        pressure=1.01325,  # bar
        path_length=1,  # cm
        verbose=3,
    )

    def compare_dataframe(df_vaex, df_pandas, columns):
        if_same = True
        for column in columns:
            if_same = if_same * np.all(df_vaex[column].to_numpy() == df_pandas[column])

        return if_same

    # HITRAN
    # Comparing Vaex dataframe and Pandas dataframe fetched from HITRAN

    # Fetching in vaex dataframe format
    config["DATAFRAME_ENGINE"] = "vaex"
    sf.fetch_databank("hitran", memory_mapping_engine="vaex", load_columns="all")
    df1 = sf.df0
    assert isinstance(df1, vaex.dataframe.DataFrameLocal)

    # Fetching in Pandas dataframe format
    config["DATAFRAME_ENGINE"] = "pandas"
    sf.fetch_databank("hitran", memory_mapping_engine="vaex", load_columns="all")
    df2 = sf.df0
    assert isinstance(df2, pandas.core.frame.DataFrame)

    # Comparing the dataframe fetched from HITEMP
    assert compare_dataframe(df1, df2, df2.columns)

    # HITEMP
    # Comparing Vaex dataframe and Pandas dataframe fetched from HITEMP

    # Fetching in vaex dataframe format
    config["DATAFRAME_ENGINE"] = "vaex"
    sf.fetch_databank("hitemp", memory_mapping_engine="vaex", load_columns="all")
    df1 = sf.df0
    assert isinstance(df1, vaex.dataframe.DataFrameLocal)

    # Fetching in Pandas dataframe format
    config["DATAFRAME_ENGINE"] = "pandas"
    sf.fetch_databank("hitemp", memory_mapping_engine="vaex", load_columns="all")
    df2 = sf.df0
    assert isinstance(df2, pandas.core.frame.DataFrame)

    # Comparing the dataframe fetched from HITEMP
    assert compare_dataframe(df1, df2, df2.columns)

    # EXOMOL
    # Comparing Vaex dataframe and Pandas dataframe fetched from EXOMOL

    # Fetching in vaex dataframe format
    config["DATAFRAME_ENGINE"] = "vaex"
    sf.fetch_databank("exomol", memory_mapping_engine="vaex", load_columns="all")
    df1 = sf.df0
    assert isinstance(df1, vaex.dataframe.DataFrameLocal)

    # Fetching in Pandas dataframe format
    config["DATAFRAME_ENGINE"] = "pandas"
    sf.fetch_databank("exomol", memory_mapping_engine="vaex", load_columns="all")
    df2 = sf.df0
    assert isinstance(df1, vaex.dataframe.DataFrameLocal)

    # Comparing the dataframe fetched from EXOMOL
    assert compare_dataframe(df1, df2, df2.columns)

    # GEISA
    # Comparing Vaex dataframe and Pandas dataframe fetched from GEISA

    # Fetching in vaex dataframe format
    config["DATAFRAME_ENGINE"] = "vaex"
    sf.fetch_databank("geisa", memory_mapping_engine="vaex", load_columns="all")
    df1 = sf.df0
    assert isinstance(df1, vaex.dataframe.DataFrameLocal)

    # Fetching in Pandas dataframe format
    config["DATAFRAME_ENGINE"] = "pandas"
    sf.fetch_databank("geisa", memory_mapping_engine="vaex", load_columns="all")
    df2 = sf.df0
    assert isinstance(df1, vaex.dataframe.DataFrameLocal)

    # Comparing the dataframe fetched from GEISA
    assert compare_dataframe(df1, df2, df2.columns)

    # Kurucz
    # Using an atom rather than molecule:
    sf = SpectrumFactory(
        1500,
        2000,
        species="C_I",
        pressure=1.01325,  # bar
        path_length=1,  # cm
    )

    config["ALLOW_OVERWRITE"] = True

    # Fetching in vaex dataframe format
    config["DATAFRAME_ENGINE"] = "vaex"
    sf.fetch_databank(
        "kurucz", memory_mapping_engine="vaex", load_columns="all", parfuncfmt="kurucz"
    )
    df1 = sf.df0
    assert isinstance(df1, vaex.dataframe.DataFrameLocal)

    # Fetching in Pandas dataframe format
    config["DATAFRAME_ENGINE"] = "pandas"
    sf.fetch_databank(
        "kurucz", memory_mapping_engine="vaex", load_columns="all", parfuncfmt="kurucz"
    )
    df2 = sf.df0
    assert isinstance(df1, vaex.dataframe.DataFrameLocal)

    # Comparing the dataframe fetched from GEISA
    assert compare_dataframe(df1, df2, df2.columns)


@pytest.mark.skipif(isinstance(vaex, NotInstalled), reason="Vaex not available")
def test_vaex_and_pandas_dataframe_load_databank():
    """
    Compares dataframes fetched and manipulated with vaex with that of pandas.
    Every column of dataframe in vaex format is compared with corresponding column
    of dataframe in pandas format .
    Added in https://github.com/radis/radis/pull/580
    """

    from radis import SpectrumFactory

    # Comparing HITRAN
    sf = SpectrumFactory(
        wavenum_min=2380,
        wavenum_max=2400,
        mole_fraction=400e-6,
        path_length=100,  # cm
        isotope=[1],
    )

    # Loading in vaex dataframe format
    config["DATAFRAME_ENGINE"] = "vaex"
    sf.load_databank("HITRAN-CO2-TEST")
    df1 = sf.df0

    # Loading in  pandas dataframe format
    config["DATAFRAME_ENGINE"] = "pandas"
    sf.load_databank("HITRAN-CO2-TEST")
    df2 = sf.df0

    # Comparing both the dataframes
    for column in df2.columns:
        assert np.all(df1[column].to_numpy() == df2[column])

    # Comparing HITEMP
    sf = SpectrumFactory(
        2284.2,
        2284.6,
        wstep=0.001,  # cm-1
        pressure=20 * 1e-3,  # bar
        cutoff=0,
        path_length=0.1,
        mole_fraction=400e-6,
        molecule="CO2",
        isotope="1,2",
        medium="vacuum",
        truncation=5,
        verbose=0,
    )
    sf.warnings["MissingSelfBroadeningWarning"] = "ignore"

    # Loading in vaex dataframe format
    config["DATAFRAME_ENGINE"] = "vaex"
    sf.load_databank("HITEMP-CO2-TEST", load_columns="all")
    df1 = sf.df0

    # Loading in pandas dataframe format
    config["DATAFRAME_ENGINE"] = "pandas"
    sf.load_databank("HITEMP-CO2-TEST", load_columns="all")
    df2 = sf.df0

    # Comparing both the dataframes
    for column in df2.columns:
        assert np.all(df1[column].to_numpy() == df2[column])


def _run_testcases(verbose=True, plot=False):

    test_retrieve_from_database(plot=plot, verbose=verbose)
    # test_ignore_cached_files()
    # test_ignore_irrelevant_files(verbose=verbose)
    # test_custom_abundance()
    # test_vaex_and_pandas_dataframe_fetch_databank()
    # test_vaex_and_pandas_dataframe_load_databank()


if __name__ == "__main__":
    _run_testcases()

# %%



# -*- coding: utf-8 -*-
"""

Summary
-----------------------

GEISA database parser

-----------------------

Defines :py:func:`~radis.io.fetch_geisa` based on :py:class:`~radis.api.geisaapi.GEISADatabaseManager`


"""

from os.path import abspath, expanduser, join

from radis.api.geisaapi import GEISADatabaseManager


def fetch_geisa(
    molecule,
    local_databases=None,
    databank_name="GEISA-{molecule}",
    isotope=None,
    load_wavenum_min=None,
    load_wavenum_max=None,
    columns=None,
    cache=True,
    verbose=True,
    chunksize=100000,
    clean_cache_files=True,
    return_local_path=False,
    engine="default",
    output="pandas",
    parallel=True,
):
    """Stream GEISA file from GEISA website. Unzip and build a HDF5 file directly.

    Returns a Pandas DataFrame containing all lines.

    Parameters
    ----------
    molecule: all 58 GEISA 2020 molecules. See here https://geisa.aeris-data.fr/interactive-access/?db=2020&info=ftp
    local_databases: str
        where to create the RADIS HDF5 files. Default ``"~/.radisdb/geisa"``.
        Can be changed in ``radis.config["DEFAULT_DOWNLOAD_PATH"]`` or in ~/radis.json config file
    databank_name: str
        name of the databank in RADIS :ref:`Configuration file <label_lbl_config_file>`
        Default ``"GEISA-{molecule}"``
    isotope: str, int or None
        load only certain isotopes : ``'2'``, ``'1,2'``, etc. If ``None``, loads
        everything. Default ``None``.
    load_wavenum_min, load_wavenum_max: float (cm-1)
        load only specific wavenumbers.
    columns: list of str
        list of columns to load. If ``None``, returns all columns in the file.

    Other Parameters
    ----------------
    cache: ``True``, ``False``, ``'regen'`` or ``'force'``
        if ``True``, use existing HDF5 file. If ``False`` or ``'regen'``, rebuild it.
        If ``'force'``, raise an error if cache file cannot be used (useful for debugging).
        Default ``True``.
    verbose: bool
    chunksize: int
        number of lines to process at a same time. Higher is usually faster
        but can create Memory problems and keep the user uninformed of the progress.
    clean_cache_files: bool
        if ``True`` clean downloaded cache files after HDF5 are created.
    return_local_path: bool
        if ``True``, also returns the path of the local database file.
    engine: 'pytables', 'vaex', 'default'
        which HDF5 library to use to parse local files. If 'default' use the value from ~/radis.json
    output: 'pandas', 'vaex', 'jax'
        format of the output DataFrame. If ``'jax'``, returns a dictionary of
        jax arrays.
    parallel: bool
        if ``True``, uses joblib.parallel to load database with multiple processes

    Returns
    -------
    df: pd.DataFrame or vaex.dataframe.DataFrameLocal
        Line list
        A HDF5 file is also created in ``local_databases`` and referenced
        in the :ref:`RADIS config file <label_lbl_config_file>` with name
        ``databank_name``
    local_path: str
        path of local database file if ``return_local_path``

    Examples
    --------
    ::

        from radis import fetch_geisa
        df = fetch_geisa("CO")
        print(df.columns)
        >>> Index(['wav', 'int', 'airbrd', 'El', 'globu', 'globl', 'locu', 'locl',
            'Tdpgair', 'isoG', 'mol', 'idG', 'id', 'iso', 'A', 'selbrd', 'Pshft',
            'Tdpair', 'ierrA', 'ierrB', 'ierrC', 'ierrF', 'ierrO', 'ierrR', 'ierrN',
            'Tdpgself', 'ierrS', 'Pshfts', 'ierrT', 'Tdppself', 'ierrU'],
            dtype='object')

    .. minigallery:: radis.fetch_geisa

    Notes
    -----
    if using ``load_only_wavenum_above/below`` or ``isotope``, the whole
    database is anyway downloaded and uncompressed to ``local_databases``
    fast access .HDF5 files (which will take a long time on first call). Only
    the expected wavenumber range & isotopes are returned. The .HFD5 parsing uses
    :py:func:`~radis.api.hdf5.hdf2df`

    See Also
    --------
    :py:func:`~radis.io.hitran.fetch_hitran`, :py:func:`~radis.io.exomol.fetch_exomol`
    :py:func:`~radis.io.hitemp.fetch_hitemp`, :py:func:`~radis.io.kurucz.fetch_kurucz`
    :py:func:`~radis.api.hdf5.hdf2df`
    :py:meth:`~radis.lbl.loader.DatabankLoader.fetch_databank`

    """

    if r"{molecule}" in databank_name:
        databank_name = databank_name.format(**{"molecule": molecule})

    if local_databases is None:
        import radis

        local_databases = join(radis.config["DEFAULT_DOWNLOAD_PATH"], "geisa")
    local_databases = abspath(expanduser(local_databases))

    ldb = GEISADatabaseManager(
        databank_name,
        molecule=molecule,
        local_databases=local_databases,
        verbose=verbose,
        chunksize=chunksize,
        parallel=parallel,
        engine=engine,
    )

    # Get list of all expected local files for this database:
    local_files, urlnames = ldb.get_filenames()

    # Delete files if needed:

    if cache == "regen":
        ldb.remove_local_files(local_files)
    ldb.check_deprecated_files(
        ldb.get_existing_files(local_files),
        auto_remove=True if cache != "force" else False,
    )

    # Get missing files
    download_files = ldb.get_missing_files(local_files)

    # Download files
    if len(download_files) > 0:
        if urlnames is None:
            urlnames = ldb.fetch_urlnames()
        filesmap = dict(zip(local_files, urlnames))
        download_urls = [filesmap[k] for k in download_files]
        ldb.download_and_parse(download_urls, download_files)

    # Register
    if not ldb.is_registered():
        ldb.register()

    if len(download_files) > 0 and clean_cache_files:
        ldb.clean_download_files()

    if isotope and type(isotope) == int:
        isotope = str(isotope)

    # Load and return
    df = ldb.load(
        local_files,  # filter other files,
        columns=columns,
        within=[("iso", isotope)] if isotope is not None else [],
        # for relevant files, get only the right range :
        lower_bound=[("wav", load_wavenum_min)] if load_wavenum_min is not None else [],
        upper_bound=[("wav", load_wavenum_max)] if load_wavenum_max is not None else [],
        output=output,
    )

    return (df, local_files) if return_local_path else df


#%%

if __name__ == "__main__":

    import pytest

    print("Testing factory:", pytest.main(["../test/io/test_geisa.py"]))



import time

import matplotlib.pyplot as plt
import numpy as np

from radis import calc_spectrum, config
from radis.misc.progress_bar import ProgressBar


def compare_vaex_pandas_time():
    """
    Compares the time performance of pandas and Vaex and generates a plot. This scripts takes several minutes to run.
    This results shoud shown that vaex and pandas provide similar performances in term if speed.
    Returns
    -------
    None.
    """
    time_list, timeC_list, lines_list = [], [], []
    time_list_va, timeC_list_va, lines_list_va = [], [], []
    wmin = 1000
    steps = 5
    wmax_arr = np.geomspace(10, 1000, steps)

    initial_engine = config[
        "DATAFRAME_ENGINE"
    ]  # To make sure dataframe engine not changed after running this test
    pb = ProgressBar(N=2 * steps)
    for i, engine in enumerate(["vaex", "pandas"]):
        config["DATAFRAME_ENGINE"] = engine
        for j, w_range in enumerate(wmax_arr):
            t0 = time.time()
            s = calc_spectrum(
                wmin,
                wmin + w_range,  # cm-1
                molecule="H2O",
                isotope="1,2,3",
                pressure=1.01325,  # bar
                Tgas=1000,
                mole_fraction=0.1,
                databank="hitemp",  # or 'hitemp'
                wstep="auto",
                cutoff=1e-28,
                verbose=0,
            )
            t1 = time.time()
            if engine == "vaex":
                timeC_list_va.append(s.conditions["calculation_time"])
                lines_list_va.append(s.conditions["lines_calculated"])
                time_list_va.append(t1 - t0)
                # lines_list_va.append(s.conditions['lines_calculated']+s.conditions['lines_cutoff'])
            else:
                timeC_list.append(s.conditions["calculation_time"])
                lines_list.append(s.conditions["lines_calculated"])
                time_list.append(t1 - t0)
                # lines_list.append(s.conditions['lines_calculated']+s.conditions['lines_cutoff'])
            pb.update(i * steps + (j + 1))
    plt.figure()
    plt.plot(lines_list, time_list, "k", label="pandas total")
    plt.plot(lines_list, timeC_list, "k--", label="pandas computation")
    plt.plot(lines_list_va, time_list_va, "r", label="vaex total")
    plt.plot(lines_list_va, timeC_list_va, "r--", label="vaex computation")
    plt.ylabel("Time [s]")
    plt.xlabel("Number of lines")
    plt.legend()

    config["DATAFRAME_ENGINE"] = initial_engine


# Compare the memory performance of Pandas and Vaex
def compare_pandas_vs_vaex_memory():
    """
    Compare memory usage of `engine="vaex"` and `engine="pandas"` in calc_spectrum.
    Expected behavior is "vaex" using much less memory. This function takes tens of seconds to run.
    Returns
    -------
    None.
    """

    import tracemalloc

    initial_engine = config[
        "DATAFRAME_ENGINE"
    ]  # To make sure dataframe engine not changed after running this test
    for engine in ["pandas", "vaex"]:
        config["DATAFRAME_ENGINE"] = engine
        tracemalloc.start()
        s = calc_spectrum(
            1000,
            1500,  # cm-1
            molecule="H2O",
            isotope="1,2,3",
            pressure=1.01325,  # bar
            Tgas=1000,  # K
            mole_fraction=0.1,
            wstep="auto",
            databank="hitemp",  # or 'hitemp', 'geisa', 'exomol'
            verbose=0,
        )
        snapshot = tracemalloc.take_snapshot()
        memory = tracemalloc.get_traced_memory()
        tracemalloc.stop()

        # Some raw outputs
        print("\n******** Engine = {} ***********".format(engine))
        print(
            "Peak, current = {:.1e}, {:.1e} for {:} lines calculated".format(
                *memory, s.conditions["lines_calculated"]
            )
        )

        # More sophisticated
        print("*** List of biggest objects ***")
        top_stats = snapshot.statistics("lineno")
        for rank, stat in enumerate(top_stats[:3]):
            print("#{}".format(rank + 1))
            print(stat)

        # Clear for next engine in the loop
        tracemalloc.clear_traces()

    config["DATAFRAME_ENGINE"] = initial_engine


compare_vaex_pandas_time()
compare_pandas_vs_vaex_memory()
